Attribute VB_Name = "Direct_Impacts_v11"
Option Explicit

'=================================================================================='
'=                                                                                ='
'=    SDC Direct Impacts v11                                                      ='
'=                                                                                ='
'=    Program to measure differences between pre- and post-perturbation counts    ='
'=                                                                                ='
'=    Author: Paul Williamson, Dept of Geography, University of Liverpool         ='
'=                                                                                ='
'=    Copyright Paul Williamson 2005                                              ='
'=                                                                                ='
'=    This program is free software; you can redistribute it and/or modify        ='
'=    it under the terms of the GNU General Public License as published by        ='
'=    the Free Software Foundation; either version 2 of the License, or           ='
'=    (at your option) any later version.                                         ='
'=                                                                                ='
'=    This program is distributed in the hope that it will be useful,             ='
'=    but WITHOUT ANY WARRANTY; without even the implied warranty of              ='
'=    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               ='
'=    GNU General Public License for more details.                                ='
'=                                                                                ='
'=    You should have received a copy of the GNU General Public License           ='
'=    along with this program; if not, write to the Free Software                 ='
'=    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA  ='
'=                                                                                ='
'=================================================================================='

Dim tablename As String
Dim NoOfTables As Integer, NoOfMaps As Integer
Dim map(1 To 30, 0 To 40) As Integer 'map of table counts on to table marginals:
                                     'mapping (1:30) by rows/cols as appropriate (0:40)
                                     'N.B. second dimension must be as large as largest
                                     'of row/column dimensions declared in other
                                     'arrays such as cell_exp
                                              
Dim NoOfPercentages, NoOfSamples As Integer
Dim MaxContribCells As Integer, MaxCellTypeII As Integer
Dim MaxContribCells_tot As Integer
Dim chiinv(0 To 5000) As Double

Dim n_valid_Celltypes As Long 'No. of cell types for given table (where cell 'type' = no. of counts contribiuting to cell count)
Dim n_cells(-2 To 50) As Long 'No. of cells of each cell type for given table (mapped)
Dim n_cells_tot(-2 To 200) 'No. of cells of each cell type across all tables (actual)
Dim n_tables_with_celltype(-2 To 200) As Integer 'No. of tables with this no. of contrib cells
Dim celltype_map(-2 To 50) As Long 'Mapping of valid cell types for given table onto celltype based arrays
Dim inv_celltype_map(-2 To 800) As Long 'mapping of celltypes onto valid cell types

Dim min_rows(1 To 20, -2 To 200) As Integer 'no. of rows of given celltype (-2:200) in each table (1:20)
Dim min_cols(1 To 20, -2 To 200) As Integer 'no. of cols of given celltype (-2:200) in each table (1:20)

Dim rows(1 To 20) As Integer, cols(1 To 20) As Integer 'no. of rows and columns in each table

'========================================================================
'=== On/Off flags
Dim strata_switch As Integer
Dim report_table_mapping As Integer
Dim report_table_totals As Integer
Dim report_table_diffs_flag As Integer, report_cell_diffs_flag As Integer
Dim report_table_diffs_header_flag As Integer
Dim report_table_summary_flag As Integer
Dim report_cell_summary_flag As Integer
Dim report_cross_table_sample_summary_flag As Integer
Dim report_cross_table_overall_summary_flag As Integer
Dim report_ranking_flag(1 To 3) As Integer
Dim Cell_Fitness_Switch(1 To 8) As Integer
Dim Table_Fitness_Switch(0 To 23) As Integer '0:23 - table fitness measures
Dim Cell_Summary_Switch(1 To 5) As Integer
Dim Table_Summary_Switch(1 To 5) As Integer
Dim Use_Percentages_Switch As Integer
Dim Use_Counts_Switch As Integer
Dim Count_table_in_use As Integer

'================================================================================
'======= Declare arrays require to cumulate sample results

' === a) Actual Table totals (based on sum of contributing internal cells; therefore
' ===                         may differ from input [perturbed] table total)
' ===    1:4 - table no.

Dim obs_table_total As Double
Dim exp_table_total As Double

' === b) Cell-level measures

'(i) table x sample x col x row
'[1:20 - table no.; 1:1000 - Sample no.; 1:20 - col. no.; 1:40 - row. no.]
Dim cell_exp(1 To 20, 1 To 1000, 1 To 20, 1 To 40) As Double 'expected (unperturbed) count
Dim cell_obs(1 To 20, 1 To 1000, 1 To 20, 1 To 40) As Double 'observed (perturbed) count

''Dim exp_pctgs(1 To 1000, 1 To 100) As Double '1:1000 - sample; 1:100 - percentage
''Dim obs_pctgs(1 To 1000, 1 To 100) As Double '1:1000 - sample; 1:100 - percentage

'(ii) NoOfSamples x col x row
'[1:1000 - Sample no.; 1:20 - col. no.; 1:40 - row. no.]
Dim cell_diff(1 To 1000, 1 To 20, 1 To 40) As Double 'Difference between cell counts (new-original)
Dim cell_changed(1 To 1000, 1 To 20, 1 To 40) As Double 'Whether cell counts differ (new=original?) (1=yes; 0=no)
Dim cell_Z(1 To 1000, 1 To 20, 1 To 40) As Double  'Cell Z-score
Dim cell_NFC(1 To 1000, 1 To 20, 1 To 40) As Double  'Whether or not cell fits (Z<+/-1.96)
Dim cell_Zm(1 To 1000, 1 To 20, 1 To 40) As Double 'Cell Z-score
Dim cell_NFCm(1 To 1000, 1 To 20, 1 To 40) As Double  'Whether or not cell fits (Z<+/-1.96)

' === c) Table-level measures
' ===     1:50 - sample no.; -2 to 50: mapped cell type [or -2 to 200: actual cell type]
' ===    [-2=marginal; -1=internal; 0 = All 1+ = no. of cells contributing to count]

Dim n_changed(1 To 1000, -2 To 50) As Double 'no. of cells in table modified by disclosure control
Dim n_changed_tot(1 To 1000, -2 To 200) As Double 'as n_changed, but summed across all tables
Dim p_changed(1 To 1000, -2 To 50) As Double '% of cells in table modified by disclosure control
Dim p_changed_tot(1 To 1000, -2 To 200) As Double 'as p_changed, but summed across all tables
Dim max_change(1 To 1000, -2 To 50) As Double 'Max change in a cell value
Dim max_change_tot(1 To 1000, -2 To 200) As Double 'as max_change, but summed across all tables
Dim max_p_change(1 To 1000, -2 To 50) As Double 'Max % change in a cell value
Dim max_p_change_tot(1 To 1000, -2 To 200) As Double 'as Max % change, but summed across all tables
Dim TE(1 To 1000, -2 To 50) As Double 'Total error
Dim TE_tot(1 To 1000, -2 To 200) As Double 'as Total error, but summed across all tables
Dim TAE(1 To 1000, -2 To 50) As Double 'Total Absolute Error
Dim TAE_tot(1 To 1000, -2 To 200) As Double 'Total Absolute Error, summed across all tables
Dim SE(1 To 1000, -2 To 50) As Double 'Square error
Dim SE_tot(1 To 1000, -2 To 200) As Double 'Square error, summed across all tables
Dim RAE(1 To 1000, -2 To 50) As Double 'Relative Absolure Error
Dim RAE_tot(1 To 1000, -2 To 200) As Double  'RAE, summed across all tables
Dim TVCC(1 To 1000, -2 To 50) 'Total Value of Changed Counts
Dim TVCC_tot(1 To 1000, -2 To 200) As Double 'As per TVCC, summed across all tables
Dim SAE(1 To 1000, -2 To 50) As Double 'Standardised Absolute Error
Dim SAE_tot(1 To 1000, -2 To 200) As Double 'Standardised Absolute Error
Dim v_expCells(1 To 1000, -2 To 50) As Double 'value of SAS counts
Dim v_expCells_tot(1 To 1000, -2 To 200) As Double 'v_expCells summed across tables
Dim v_obsCells(1 To 1000, -2 To 50) As Double 'value of perturbed counts
Dim v_obsCells_tot(1 To 1000, -2 To 200) As Double 'v_obsCells summed across tables
Dim NFC(1 To 1000, -2 To 50) As Double 'No. of non-fitting cells  [Z>+/-1.96]
Dim NFC_tot(1 To 1000, -2 To 200) As Double 'NFC summed across all tables
Dim NFT(1 To 1000, -2 To 50) As Double 'Whether or not table is non-fitting (based on SSZ)
Dim NFT_tot(1 To 1000, -2 To 200) As Double 'NFT summed across all tables
Dim SSZ(1 To 1000, -2 To 50) As Double 'Sum of Squared Zm-scores
Dim SSZ_tot(1 To 1000, -2 To 200) As Double 'SSZ summed across tables
Dim NFCm(1 To 1000, -2 To 50) As Double 'No. of non-fitting cells  [Zm>+/-1.96]
Dim NFCm_tot(1 To 1000, -2 To 200) As Double 'NFCm summed across all tables
Dim NFTm(1 To 1000, -2 To 50) As Double 'Whether or not table is non-fitting (based on SSZm)
Dim NFTm_tot(1 To 1000, -2 To 200) As Double 'NFTm summed across all tables
Dim SSZm(1 To 1000, -2 To 50) As Double 'Sum of Squared Z-scores
Dim SSZm_tot(1 To 1000, -2 To 200) As Double 'SSZm summed across all tables
Dim RMSE(1 To 1000, -2 To 50) As Double 'Root Mean Square Error
Dim RMSE_tot(1 To 1000, -2 To 200) As Double 'RMSE summed acros tables
Dim D(1 To 1000, -2 To 50) As Double 'Gibson's dissimilarity index
Dim D_tot(1 To 1000, -2 To 200) As Double 'Gibson's D summed across all tables
Dim ChiSquare(1 To 1000, -2 To 50) As Double 'Chi-Square
Dim ChiSquare_tot(1 To 1000, -2 To 200) As Double 'Chi-Square
Dim CramersV(1 To 1000, -2 To 50) As Double 'Cramer's V
Dim CramersV_tot(1 To 1000, -2 To 200) As Double 'Cramer's V
Dim PearsonsR(1 To 1000, -2 To 50) As Double 'Pearson's R
Dim PearsonsR_tot(1 To 1000, -2 To 200) As Double 'Pearson's R summed across all tables

Dim Cell_Fitness_label(1 To 8) As String
Dim Fitness_label(1 To 23) As String, t_dp(1 To 23) As Integer

Dim StrataDataPath As String, TableMappingsPath As String
Dim InputCountsPath As String, RunParametersPath As String
Dim PathType As String, ProgramPath As String

Sub main()

'File handles
Dim temp As Integer, results As Integer
Dim file, mapfile As Integer  'input file channels for table and table map data
Dim original_data As Integer, perturbed_data As Integer
Dim filelist As Integer 'input channel for list of tables to be processed
Dim sdc_original As Integer, sdc_perturbed As Integer
Dim map_filename As String

'Run-specific parameters
Dim data_source As String
Dim strata_type As String, strata_name As String
Dim strata_loop, strata As Integer
Dim sample_size As Long, sample_type As Integer, sample_type_name As String

Dim ED As Integer, Current_ED As Integer, sample As Integer
Dim x As Integer, y As Integer, line_end As Integer

Dim table_no As Integer
Dim TableTotalX As Integer 'X-coord for location of total for table T
Dim TableTotalY As Integer 'Y-coord for location of total for table T
Dim percentage_name(1 To 20, 1 To 40) As String 'percentage names
                                                '(treating percentage mappings as
                                                'regular tables with max 20 cols/40 rows
  
Dim contrib_cells(1 To 20, 1 To 20, 1 To 40) As Integer 'no. of cells contributing to each marginal
                                                        '1:20 - tables; 1:20 - x; 1:40 - y

Dim no_of_strata As Integer


'Read in input/output path names

'ProgramPath = CurDir()
ProgramPath = "C:\Temp\Test SDCi"

temp = FreeFile
Open Trim(ProgramPath) & "\" & "Input_and_output_paths.txt" For Input As #temp

Input #temp, PathType, InputCountsPath
Input #temp, PathType, TableMappingsPath
Input #temp, PathType, StrataDataPath
Input #temp, PathType, RunParametersPath

Close #temp

'StrataDataPath = "d:\research\disclosure control\disclosure_vb\SAS_Inputs\"
'TableMappingsPath = "d:\research\disclosure control\disclosure_vb\Perturb_v3\"
'RunParametersPath = "d:\research\disclosure control\disclosure_vb\SDC_Direct_Impacts_v10\"
'InputCountsPath = "d:\research\disclosure control\disclosure_vb\Create_Aggregates_v2\Output\"

'Open input and output files required throughout length of run
Call open_general_files(temp)

'Read in user-defined run-specific parameters [shared with Create_Aggregates]
Call read_run_parameters(data_source, NoOfSamples, sample_type, sample_type_name, sample_size, _
  strata_switch, report_table_mapping, _
  Use_Percentages_Switch, no_of_strata, strata_name, temp)

'Set internal run-specific parameters (not directly modifiable by external user),
Call set_internal_parameters

If Use_Counts_Switch = 1 Then

  Count_table_in_use = 1

  no_of_strata = strata_switch
  If (data_source <> "Create_Aggregates") Then no_of_strata = 1

  For strata_loop = 1 To no_of_strata

    'Given strata_switch and strata_loop no., determine current strata type
  
    If strata_switch = 1 Then
      strata = 0
    ElseIf strata_switch = 2 Then
      If strata_loop = 1 Then
        strata = 1
      Else
        strata = 2
      End If
    ElseIf strata_switch = 3 Then
      If strata_loop = 1 Then
        strata = 0
      ElseIf strata_loop = 2 Then
        strata = 1
      ElseIf strata_loop = 3 Then
        strata = 2
      End If
    End If
    
    'Identify strata type (for use in file extensions)
    If strata = 0 Then
      strata_type = "_All"
    ElseIf strata = 1 Then
      strata_type = "_P20"
    ElseIf strata = 2 Then
      strata_type = "_P80"
    End If
           
    'Print #temp, "+++ Strata Type: "; strata_type; " (Strata loop "; strata_loop; ") +++"
    'Print #temp,
      
    'Reset strata-specific variables to zero
    Call reset_strata_specific_arrays(table_no)

    'Open file containing list of tables to be processed
    filelist = FreeFile
    Open ProgramPath & RunParametersPath & "SDC_Direct_Impacts_Count_input_tables.txt" For Input As #filelist

    Do While Not EOF(filelist)
    
      Input #filelist, tablename, sdc_original, sdc_perturbed
      map_filename = Trim(tablename) & ".map"
      'print #temp, Trim(tablename),sdc_orginal, sdc_perturbed, map_filename
  
      If (data_source = "Create_Aggregates") Then
        Print #temp, "=== Strata"; strata; "[strata_type "; strata_type; "] "; tablename; " ==="
      Else
        Print #temp, "=== User-supplied data: "; tablename; " ==="
      End If
      Print #temp,
  
      table_no = table_no + 1
  
      NoOfTables = NoOfTables + 1

      'Open table-specific files
      Call open_table_specific_files(data_source, tablename, sdc_original, sdc_perturbed, _
        strata_type, strata_name, sample_size, sample_type_name, NoOfSamples, map_filename, _
        original_data, perturbed_data, mapfile, temp)
  
      ' Read in table mappings
      Call table_mappings(mapfile, temp, table_no, contrib_cells, _
        TableTotalX, TableTotalY)

      'Reset table-specific (array) variable values to 0 as required
      Call reset_table_specific_arrays(table_no)

      'Scroll through samples one at a time (listed in *****)

      For sample = 1 To NoOfSamples 'MUST = NoOfSamples

        'For each sample set

  
        '(1) Read in aggregated data for each sample, for each table

         Call get_sample_tables(table_no, sample, _
           TableTotalX, TableTotalY, original_data, perturbed_data, temp)

         '(2)Measure size of perturbation (for each cumulated sample table count and marginal)
         '    and save for analysis of between sample variability

         Call Measure_diffs(table_no, TableTotalX, TableTotalY, _
           contrib_cells, temp, sample)

      Next sample 'next sample

      If report_table_diffs_flag = 1 Then Print #temp, 'add blank line to output

      If (report_cell_summary_flag = 1) Then Call Summarize_Cellular_Results(table_no, temp)

      If (report_table_summary_flag = 1) Then Call Summarize_Tabular_Results(temp)

      'close table-specific files
      Close #original_data
      Close #perturbed_data
      Close #mapfile

    Loop 'next table

    'Calculate and report cross-table summaries only if required
    If (report_cross_table_sample_summary_flag = 1) Or _
      (report_cross_table_overall_summary_flag = 1) Then
    
      Call Calc_Crosstable_Diffs(contrib_cells)

      If (report_cross_table_sample_summary_flag = 1) Then _
      Call Report_Cross_Table_Sample_Summary(temp) '[if required]

      If (report_cross_table_overall_summary_flag = 1) Then _
        Call Report_Cross_Table_Overall_Summary(temp)

    End If 'if cross-table summaries required

    Next strata_loop
  
End If 'if Use_Counts_Switch=1
  
'================= PERCENTAGES ==============================

If (Use_Percentages_Switch = 1) Or (Use_Percentages_Switch = 2) Then

Print #temp, "===================== PERCENTAGES ==================="
Print #temp,
Print #temp,

Count_table_in_use = 0

'Amend fitness flags for perecentage data (if required)
Call amend_GoF_flags(temp)



no_of_strata = strata_switch
If (data_source <> "Create_Aggregates") Then no_of_strata = 1

For strata_loop = 1 To no_of_strata

  'Given strata_switch and strata_loop no., determine strata type
  
  If strata_switch = 1 Then
    strata = 0
  ElseIf strata_switch = 2 Then
    If strata_loop = 1 Then
      strata = 1
    Else
      strata = 2
    End If
  ElseIf strata_switch = 3 Then
    If strata_loop = 1 Then
      strata = 0
    ElseIf strata_loop = 2 Then
      strata = 1
    ElseIf strata_loop = 3 Then
      strata = 2
    End If
  End If
    
  'Identify strata type (for use in file extensions)
  If strata = 0 Then
    strata_type = "_All"
  ElseIf strata = 1 Then
    strata_type = "_P20"
  ElseIf strata = 2 Then
    strata_type = "_P80"
  End If
        
  'Print #temp, "+++ Strata Type: "; strata_type; " (Strata loop "; strata_loop; ") +++"
  'Print #temp,
       
  'Reset strata-specific variables to zero
  Call reset_strata_specific_arrays(table_no)

  'Open file containing list of tables to be processed
  filelist = FreeFile
  Open ProgramPath & RunParametersPath & "SDC_Direct_Impacts_Percentage_input_tables.txt" For Input As #filelist
  
  Do While Not EOF(filelist)
    
    Input #filelist, tablename, sdc_original, sdc_perturbed
    map_filename = Trim(tablename) & ".map"
    'print #temp, Trim(tablename),sdc_orginal, sdc_perturbed , map_filename
  
  
    If (data_source = "Create_Aggregates") Then
      Print #temp, "=== Strata"; strata; "[strata_type "; strata_type; "] "; tablename; " ==="
    Else
      Print #temp, "=== User-supplied data: "; tablename; " ==="
    End If
    Print #temp,
  
    table_no = table_no + 1
  
    NoOfTables = NoOfTables + 1

    'Open percentage-specific files
    Call open_percentage_specific_files(data_source, tablename, sdc_original, sdc_perturbed, _
      strata_type, strata_name, sample_size, sample_type_name, NoOfSamples, map_filename, _
      original_data, perturbed_data, mapfile, temp)
  
    ' Read in table mappings
    Call table_mappings(mapfile, temp, table_no, contrib_cells, _
      TableTotalX, TableTotalY)
    
    'Reset table-specific (array) variable values to 0 as required
    Call reset_table_specific_arrays(table_no)
        
    'Read in percentage names

    For y = 1 To rows(table_no)
      For x = 1 To cols(table_no)
        Input #original_data, percentage_name(x, y)
        Input #perturbed_data, percentage_name(x, y)
        Print #temp, "Percentage "; x; " name: "; percentage_name(x, y)
      Next x
    Next y
    Input #original_data, line_end
    Input #perturbed_data, line_end
    Print #temp,
    
    'Scroll through samples one at a time

    For sample = 1 To NoOfSamples 'MUST = NoOfSamples

      'For each sample set

      '(1) Read in aggregated data for each sample, for each table

       Call get_sample_tables(table_no, sample, _
         TableTotalX, TableTotalY, original_data, perturbed_data, temp)

       '(2)Measure size of perturbation (for each cumulated sample table count and marginal)
       '    and save for analysis of between sample variability

       Call Measure_diffs(table_no, TableTotalX, TableTotalY, _
         contrib_cells, temp, sample)

    Next sample 'next sample

    If report_table_diffs_flag = 1 Then Print #temp, 'add blank line to output

    If (report_cell_summary_flag = 1) Then Call Summarize_Cellular_Results(table_no, temp)

    If (report_table_summary_flag = 1) Then Call Summarize_Tabular_Results(temp)

    
    'Check impact of SDC on rankings.  Subroutine requires two numeric inputs:
    '(i) # of classes to be used;
    '(ii) 0/1 flag indicating whether or not to print ranking info. (only required for first class)
    Call check_rankings(table_no, percentage_name, 5, 1, temp)
    Call check_rankings(table_no, percentage_name, 10, 0, temp)
    Call check_rankings(table_no, percentage_name, 20, 0, temp)
    
    'close table-specific files
    Close #original_data
    Close #perturbed_data
    Close #mapfile

  Loop 'next table

  'Calculate and report cross-table summaries only if required
  If (report_cross_table_sample_summary_flag = 1) Or _
    (report_cross_table_overall_summary_flag = 1) Then
    
    Call Calc_Crosstable_Diffs(contrib_cells)

    If (report_cross_table_sample_summary_flag = 1) Then _
      Call Report_Cross_Table_Sample_Summary(temp)
  
    If (report_cross_table_overall_summary_flag = 1) Then _
      Call Report_Cross_Table_Overall_Summary(temp)
  
  End If 'if cross-table reports required


Next strata_loop
  
End If 'if percentages_switch = 1 [on]

Close #temp

End Sub
Sub read_run_parameters(data_source, NoOfSamples, sample_type, sample_type_name, sample_size, _
  strata_switch, report_table_mapping, _
  Use_Percentages_Switch, no_of_strata, strata_name, temp)

Dim parameters_in As Integer, strata_in As Integer
Dim label As String, strata_source_file As String

Dim total_cell_fitness As Integer, total_cell_summary As Integer, i As Integer
Dim total_table_fitness As Integer, total_table_summary As Integer

'Open file containing run-specific parameters
parameters_in = FreeFile
Open ProgramPath & RunParametersPath & "SDC_Direct_Impacts_run_parameters.txt" For Input As #parameters_in

'=== Set sample information

Print #temp, "=== User-defined run parameters ==="
Print #temp,

'Read in/write out section header
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)

'(a) Data source for input orginal/perturbed counts
Input #parameters_in, label, data_source
Print #temp, Trim(label); " "; data_source

If (data_source <> "Create_Aggregates") And (data_source <> "User") Then
  MsgBox ("Error: data source <> ""Create_Aggregates"" or ""User""; program execution terminated.")
  Stop
End If

'(a) No. of samples

'NoOfSamples must <= NoOfValidEDs if each ED to be processed in turn;
'Max = 1000

Input #parameters_in, label, NoOfSamples
Print #temp, Trim(label); NoOfSamples

'(b) Stratification switch setting [0 = None; 1 = P20/P80/ 2 = None/P20/P80]
'[If switched on, final part of this sub-routine reads in name of source_file for strata]
Input #parameters_in, label, strata_switch
Select Case strata_switch
Case 1
  Print #temp, Trim(label); strata_switch; " [No strata]"
Case 2
  Print #temp, Trim(label); strata_switch; " [Lower and upper quintile strata]"
Case 3
  Print #temp, Trim(label); strata_switch; " [No Strata; Lower and upper quintile strata]"
End Select

'(c) Sample type [sampling strategy]

'[1=each ED in turn; 2= consecutive EDs; 3 = random EDs];
'*** Option 1 does not work with strata ***

Input #parameters_in, label, sample_type
Print #temp, Trim(label); sample_type; Trim(sample_type_name);
Select Case sample_type
  Case 1
    sample_type_name = "[A]"
    Print #temp, Trim(sample_type_name); _
      " [Each ED in turn (no aggregation)]"
  Case 2
    sample_type_name = "[C]"
    Print #temp, Trim(sample_type_name); _
      " [Consecutive EDs (aggregation)]"
  Case 3
    sample_type_name = "[R]"
    Print #temp, Trim(sample_type_name); _
      " [Random EDs (aggregation)]"
  Case 4
    sample_type_name = "[G]"
    Print #temp, Trim(sample_type_name); _
      " [Consecutive EDs within Geographical Cluster(aggregation)]"
End Select

'(d) Sample size (no. of areas to be aggregated per sample)

Input #parameters_in, label, sample_size
Print #temp, Trim(label); sample_size, "[i.e. no. of areas aggregated together per sample]"


'==== Set other run-specific parameters

'(a) Table mapping switch
Input #parameters_in, label, report_table_mapping
Print #temp, Trim(label); report_table_mapping; " [1=on; else = off]"

'(b) Percentages switch
Input #parameters_in, label, Use_Percentages_Switch
Print #temp, Trim(label); Use_Percentages_Switch

Use_Counts_Switch = 1
If Use_Percentages_Switch = 1 Then Use_Counts_Switch = 0

'==== Read in name of source file for strata data (if required);
'then, if appropriate, read in strata name from strata source file
  
Input #parameters_in, label, strata_source_file
If strata_switch = 1 Then strata_source_file = ""
  
Print #temp, "strata_switch: "; strata_switch
Print #temp, Trim(label); " "; strata_source_file

If strata_switch <> 1 Then
  strata_in = FreeFile
  Open ProgramPath & StrataDataPath & Trim(strata_source_file) For Input As #strata_in
  Input #strata_in, label, strata_name
  Print #temp, "Strata_name: "; strata_name
  Close #strata_in
End If


'Read in/write out section header
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)

'Report table totals switch [1='on';0='off']
Input #parameters_in, label, report_table_totals
'Print #temp, Trim(label); " "; report_table_totals

'Read in and discard blank line
Input #parameters_in, label

'Initialise table-specific report switches [1='on';0='off']
Input #parameters_in, label, report_cell_diffs_flag
'Print #temp, Trim(label); " "; report_cell_diffs_flag
Input #parameters_in, label, report_table_diffs_flag
'Print #temp, Trim(label); " "; report_table_diffs_flag
Input #parameters_in, label, report_cell_summary_flag
'Print #temp, Trim(label); " "; report_cell_summary_flag
Input #parameters_in, label, report_table_summary_flag
'Print #temp, Trim(label); " "; report_table_summary_flag

'Initialise cross-table summary report switches [1='on';0='off']
Input #parameters_in, label, report_cross_table_sample_summary_flag
'Print #temp, Trim(label); " "; report_cross_table_sample_summary_flag
Input #parameters_in, label, report_cross_table_overall_summary_flag
'Print #temp, Trim(label); " "; report_cross_table_overall_summary_flag

'Read in and discard blank line
Input #parameters_in, label

'Ranking report flags [1='on';0='off']
Input #parameters_in, label, report_ranking_flag(1)
'Print #temp, Trim(label); " "; report_ranking_flag(1)
Input #parameters_in, label, report_ranking_flag(2)
'Print #temp, Trim(label); " "; report_ranking_flag(2)
Input #parameters_in, label, report_ranking_flag(3)
'Print #temp, Trim(label); " "; report_ranking_flag(3)


'Read in/write out section header
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)

'Initialise output Fitness measure switches [1='on'; 0='off']
Input #parameters_in, label, Cell_Fitness_Switch(1)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(1)
Input #parameters_in, label, Cell_Fitness_Switch(2)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(2)
Input #parameters_in, label, Cell_Fitness_Switch(3)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(3)
Input #parameters_in, label, Cell_Fitness_Switch(4)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(4)
Input #parameters_in, label, Cell_Fitness_Switch(5)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(5)
Input #parameters_in, label, Cell_Fitness_Switch(6)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(6)
Input #parameters_in, label, Cell_Fitness_Switch(7)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(7)
Input #parameters_in, label, Cell_Fitness_Switch(8)
'Print #temp, Trim(label); " "; Cell_Fitness_Switch(8)

total_cell_fitness = 0
For i = 1 To 8
  total_cell_fitness = total_cell_fitness + Cell_Fitness_Switch(i)
Next i

'Read in/write out blank line
Input #parameters_in, label
'Print #temp, Trim(label)

'Initialise output summary measure switches  [1='on'; 0='off']
Input #parameters_in, label, Cell_Summary_Switch(1)
'Print #temp, Trim(label); " "; Cell_Summary_Switch(1)
Input #parameters_in, label, Cell_Summary_Switch(2)
'Print #temp, Trim(label); " "; Cell_Summary_Switch(2)
Input #parameters_in, label, Cell_Summary_Switch(3)
'Print #temp, Trim(label); " "; Cell_Summary_Switch(3)
Input #parameters_in, label, Cell_Summary_Switch(4)
'Print #temp, Trim(label); " "; Cell_Summary_Switch(4)
Input #parameters_in, label, Cell_Summary_Switch(5)
'Print #temp, Trim(label); " "; Cell_Summary_Switch(5)

total_cell_summary = 0
For i = 1 To 5
  total_cell_summary = total_cell_summary + Cell_Summary_Switch(i)
Next i

'Check to make sure that if table_diffs OR cross_table_overall_summary report
'is required, at least one table GoF measure AND one table summary type is switched on
'[else halt program execution]

If (report_cell_diffs_flag = 1 And total_cell_fitness = 0) Then
    MsgBox ("Area-specific cell differences report requires at least one cell-based measure of fit to be switched on; program execution terminated.")
    Stop
End If

'Read in/write out section header
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)
Input #parameters_in, label
'Print #temp, Trim(label)

Input #parameters_in, label, Table_Fitness_Switch(0)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(0)
Input #parameters_in, label, Table_Fitness_Switch(1)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(1)
Input #parameters_in, label, Table_Fitness_Switch(2)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(2)
Input #parameters_in, label, Table_Fitness_Switch(3)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(3)
Input #parameters_in, label, Table_Fitness_Switch(4)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(4)
Input #parameters_in, label, Table_Fitness_Switch(5)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(5)
Input #parameters_in, label, Table_Fitness_Switch(6)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(6)
Input #parameters_in, label, Table_Fitness_Switch(7)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(7)
Input #parameters_in, label, Table_Fitness_Switch(8)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(8)
Input #parameters_in, label, Table_Fitness_Switch(9)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(9)
Input #parameters_in, label, Table_Fitness_Switch(10)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(10)
Input #parameters_in, label, Table_Fitness_Switch(11)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(11)
Input #parameters_in, label, Table_Fitness_Switch(12)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(12)
Input #parameters_in, label, Table_Fitness_Switch(13)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(13)
Input #parameters_in, label, Table_Fitness_Switch(14)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(14)
Input #parameters_in, label, Table_Fitness_Switch(15)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(15)
Input #parameters_in, label, Table_Fitness_Switch(16)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(16)
Input #parameters_in, label, Table_Fitness_Switch(17)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(17)
Input #parameters_in, label, Table_Fitness_Switch(18)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(18)
Input #parameters_in, label, Table_Fitness_Switch(19)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(19)
Input #parameters_in, label, Table_Fitness_Switch(20)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(20)
Input #parameters_in, label, Table_Fitness_Switch(21)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(21)
Input #parameters_in, label, Table_Fitness_Switch(22)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(22)
Input #parameters_in, label, Table_Fitness_Switch(23)
'Print #temp, Trim(label); " "; Table_Fitness_Switch(23)

total_table_fitness = 0
For i = 0 To 23
  total_table_fitness = total_table_fitness + Table_Fitness_Switch(i)
Next i

'Read in/write out blank line
Input #parameters_in, label
'Print #temp, Trim(label)

Input #parameters_in, label, Table_Summary_Switch(1)
'Print #temp, Trim(label); " "; Table_Summary_Switch(1)
Input #parameters_in, label, Table_Summary_Switch(2)
'Print #temp, Trim(label); " "; Table_Summary_Switch(2)
Input #parameters_in, label, Table_Summary_Switch(3)
'Print #temp, Trim(label); " "; Table_Summary_Switch(3)
Input #parameters_in, label, Table_Summary_Switch(4)
'Print #temp, Trim(label); " "; Table_Summary_Switch(4)
Input #parameters_in, label, Table_Summary_Switch(5)
'Print #temp, Trim(label); " "; Table_Summary_Switch(5)

total_table_summary = 0
For i = 1 To 5
  total_table_summary = total_table_summary + Table_Summary_Switch(i)
Next i

'Check to make sure that if table_diffs report, OR cross_table_sample_summary report,
'OR cross_table_overall_summary report is required, than at least ONE
'table GoF measure AND one table summary type is switched on
'[else halt program execution]

If (report_table_diffs_flag = 1 And total_table_fitness = 0) Then
    MsgBox ("Area-specific report of tabular differences requires at least one table-based measure of fit to be switched on; program execution terminated.")
    Stop
End If

If (report_cross_table_sample_summary_flag = 1 And total_table_fitness = 0) Then
    MsgBox ("Cross-table sample summary results requires at least one table-based measure of fit to be switched on; program execution terminated.")
    Stop
End If


'    Or report_cross_table_overall_summary_flag = 1) And _

'Add blank line after parameter reports
Print #temp,

Close #parameters_in

End Sub
Sub set_internal_parameters()
Dim chisq_data As Integer, df As Integer, j As Integer
Dim label As String

NoOfTables = 0
NoOfPercentages = 0

'Set value of inverse chi-square array for degrees of freedom 1 to 1000, p=0.05
chisq_data = FreeFile
Open ProgramPath & RunParametersPath & "chisquare.dat" For Input As #chisq_data
chiinv(0) = 0# 'Needed for tables with no marginal cells
For df = 1 To 5000
  Input #chisq_data, j, chiinv(df)
Next df
Close #chisq_data

'Initialise fitness labels
Cell_Fitness_label(1) = "original_cnt"
Cell_Fitness_label(2) = "perturbed_ct"
Cell_Fitness_label(3) = "cell_changed"
Cell_Fitness_label(4) = "cell_error  "
Cell_Fitness_label(5) = "cell_Z      "
Cell_Fitness_label(6) = "cell_NFC    "
Cell_Fitness_label(7) = "cell_Zm     "
Cell_Fitness_label(8) = "cell_NFCm   "

'Initialise fitness labels
Fitness_label(1) = "n_changed "
Fitness_label(2) = "p_changed "
Fitness_label(3) = "max_change"
Fitness_label(4) = "maxPchange"
Fitness_label(5) = "TotalError"
Fitness_label(6) = "TAE       "
Fitness_label(7) = "RAE       "
Fitness_label(8) = "SAE       "
Fitness_label(9) = "Sq_Error  "
Fitness_label(10) = "RMSE      "
Fitness_label(11) = "SSZ       "
Fitness_label(12) = "NFC       "
Fitness_label(13) = "NFT       "
Fitness_label(14) = "SSZm      "
Fitness_label(15) = "NFCm      "
Fitness_label(16) = "NFTm      "
Fitness_label(17) = "Gibsons_D "
Fitness_label(18) = "Cramers_V "
Fitness_label(19) = "PearsonsR "
Fitness_label(20) = "ChiSquare "
Fitness_label(21) = "TVCC      "
Fitness_label(22) = "v_expCells"
Fitness_label(23) = "v_obsCells"

'Initialise printing decimal places for measure
t_dp(1) = 6 'n_changed
t_dp(2) = 6 'p_changed
t_dp(3) = 6 'max_change
t_dp(4) = 6 'maxPchange
t_dp(5) = 5 'TotalError
t_dp(6) = 6 'TAE
t_dp(7) = 6 'RAE
t_dp(8) = 6 'SAE
t_dp(9) = 4 'Sq_Error
t_dp(10) = 6 'RMSE
t_dp(11) = 6 'SSZ
t_dp(12) = 6 'NFC
t_dp(13) = 6 'NFT
t_dp(14) = 6 'SSZm
t_dp(15) = 6 'NFCm
t_dp(16) = 6 'NFTm
t_dp(17) = 6 'Gibsons_D
t_dp(18) = 6 'Cramers_V
t_dp(19) = 6 'PearsonsR
t_dp(20) = 6 'ChiSquare
t_dp(21) = 2 'TVCC
t_dp(22) = 2 'v_expCells
t_dp(23) = 2 'v_obsCells

End Sub
Sub amend_GoF_flags(temp)
Dim x As Integer, need_blank_line As Integer

need_blank_line = 0

'Reset output Fitness measure switches to 0 ['off'] if inappropriate for percentages

'(a) Cell level
For x = 5 To 8 'cell_Z; cell_NFC; cell_Zm; cell_NFCm
If Cell_Fitness_Switch(x) = 1 Then
  Cell_Fitness_Switch(x) = 0
  Print #temp, Cell_Fitness_label(x) & " inappropriate for percentage data, so switched off"
  need_blank_line = 1
End If
Next x

'(b) Table level
For x = 11 To 16 'SSZ; NFC; NFT; SSZm; NFCm; NFTm
  If Table_Fitness_Switch(x) = 1 Then
    Table_Fitness_Switch(x) = 0
  Print #temp, Fitness_label(x) & " inappropriate for percentage data, so switched off"
  need_blank_line = 1
  End If
Next x
 
If Table_Fitness_Switch(18) = 1 Then 'Cramers_V
  Table_Fitness_Switch(18) = 0
  Print #temp, Fitness_label(18) & " inappropriate for percentage data, so switched off"
  need_blank_line = 1
End If

If Table_Fitness_Switch(20) = 1 Then 'Chi-Square
  Table_Fitness_Switch(20) = 0
  Print #temp, Fitness_label(20) & " inappropriate for percentage data, so switched off"
  need_blank_line = 1
End If

If need_blank_line = 1 Then Print #temp,

End Sub

Sub reset_strata_specific_arrays(table_no)

Dim s As Integer, x As Integer

table_no = 0

MaxContribCells_tot = 0

NoOfTables = 0

For x = -2 To 200
  
    n_cells_tot(x) = 0#
    n_tables_with_celltype(x) = 0#
  
  For s = 1 To NoOfSamples

    n_changed_tot(s, x) = 0#
    max_change_tot(s, x) = 0#
    max_p_change_tot(s, x) = 0#
    TE_tot(s, x) = 0#
    TAE_tot(s, x) = 0#
    SE_tot(s, x) = 0#
    TVCC_tot(s, x) = 0#
    If (x <= 50) Then v_expCells(s, x) = 0#
    If (x <= 50) Then v_obsCells(s, x) = 0#
    NFC_tot(s, x) = 0#
    NFT_tot(s, x) = 0#
    SSZ_tot(s, x) = 0#
    NFCm_tot(s, x) = 0#
    NFTm_tot(s, x) = 0#
    SSZm_tot(s, x) = 0#
    ChiSquare_tot(s, x) = 0#
    CramersV_tot(s, x) = 0#
  Next s
  
Next x

End Sub
Sub reset_table_specific_arrays(table_no)

Dim s As Integer, x As Integer, y As Integer

report_table_diffs_header_flag = 1

For s = 1 To NoOfSamples

  'Reset tabel cell stats arrays
  For x = 1 To cols(table_no)
    For y = 1 To rows(table_no)
      cell_changed(s, x, y) = 0#
    Next y
  Next x
  
  'Reset table summary stats arrays
  For x = -2 To n_valid_Celltypes
      NFC(s, x) = 0#
      NFCm(s, x) = 0#
      n_changed(s, x) = 0#
      TE(s, x) = 0#
      TAE(s, x) = 0#
      TVCC(s, x) = 0#
      v_expCells(s, x) = 0#
      v_obsCells(s, x) = 0#
      SE(s, x) = 0#
      max_change(s, x) = 0#
      max_p_change(s, x) = 0#
      SSZ(s, x) = 0#
      SSZm(s, x) = 0#
      ChiSquare(s, x) = 0#
  Next x
Next s



End Sub

Sub open_general_files(temp)
  
temp = FreeFile
Open ProgramPath & "\" & "SDC_Direct_Impacts_results.txt" For Output As #temp

End Sub

Sub open_table_specific_files(data_source, tablename, sdc_original, sdc_perturbed, _
  strata_type, strata_name, sample_size, sample_type_name, NoOfSamples, map_filename, _
  original_data, perturbed_data, mapfile, temp)

Dim original_filename As String, perturbed_filename As String
Dim zoneid As String

'Original counts from >= threshold SAS/LBS tables
original_filename = Trim(tablename) & "_v" & sdc_original & _
  Trim(strata_type) & "[" & Trim(strata_name) & "]" & _
  "_n" & sample_size & Trim(sample_type_name) & "_s" & NoOfSamples & ".fmt"
If (data_source <> "Create_Aggregates") Then _
  original_filename = Trim(tablename) & "_v" & sdc_original & ".fmt"
Print #temp, "original_data: "
Print #temp, Trim(original_filename)
Print #temp, "IN: "; InputCountsPath
    
original_data = FreeFile
Open ProgramPath & InputCountsPath & Trim(original_filename) For Input As #original_data

'Perturbed version of orginal counts
perturbed_filename = Trim(tablename) & "_v" & sdc_perturbed & _
  Trim(strata_type) & "[" & Trim(strata_name) & "]" & _
  "_n" & sample_size & Trim(sample_type_name) & "_s" & NoOfSamples & ".fmt"
If (data_source <> "Create_Aggregates") Then _
  perturbed_filename = Trim(tablename) & "_v" & sdc_perturbed & ".fmt"
Print #temp, "perturbed_data: "
Print #temp, Trim(perturbed_filename)
Print #temp, "IN: "; InputCountsPath
Print #temp,
    
perturbed_data = FreeFile
Open ProgramPath & InputCountsPath & Trim(perturbed_filename) For Input As #perturbed_data

'Row/column marginal mappings
mapfile = FreeFile
Open ProgramPath & TableMappingsPath & Trim(map_filename) For Input As #mapfile


End Sub

Sub open_percentage_specific_files(data_source, tablename, sdc_original, sdc_perturbed, _
  strata_type, strata_name, sample_size, sample_type_name, NoOfSamples, map_filename, _
  original_data, perturbed_data, mapfile, temp)

Dim original_filename As String, perturbed_filename As String
Dim zoneid As String

'Original counts from >= threshold SAS/LBS tables
original_filename = Trim(tablename) & "_v" & sdc_original & _
  Trim(strata_type) & "[" & Trim(strata_name) & "]" & _
  "_n" & sample_size & Trim(sample_type_name) & "_s" & NoOfSamples & ".fmt"
If (data_source <> "Create_Aggregates") Then _
  original_filename = Trim(tablename) & "_v" & sdc_original & ".fmt"
Print #temp, "original_data: "
Print #temp, Trim(original_filename)
Print #temp, "IN: "; InputCountsPath
    
original_data = FreeFile
Open ProgramPath & InputCountsPath & Trim(original_filename) For Input As #original_data

'Perturbed version of orginal counts
perturbed_filename = Trim(tablename) & "_v" & sdc_perturbed & _
  Trim(strata_type) & "[" & Trim(strata_name) & "]" & _
  "_n" & sample_size & Trim(sample_type_name) & "_s" & NoOfSamples & ".fmt"
If (data_source <> "Create_Aggregates") Then _
  perturbed_filename = Trim(tablename) & "_v" & sdc_perturbed & ".fmt"
Print #temp, "perturbed_data: "
Print #temp, Trim(perturbed_filename)
Print #temp, "IN: "; InputCountsPath
Print #temp,
    
perturbed_data = FreeFile
Open ProgramPath & InputCountsPath & Trim(perturbed_filename) For Input As #perturbed_data

'Row/column marginal mappings
mapfile = FreeFile
Open ProgramPath & TableMappingsPath & Trim(map_filename) For Input As #mapfile


End Sub

Sub table_mappings(mapfile, temp, table_no, contrib_cells, _
  TableTotalX, TableTotalY)
Dim T As Integer, mapping As Integer, rowcol As Integer, x As Integer, y As Integer
Dim marginal As Integer, origvalue As Integer, max_contrib As Integer
Dim celltype As Long, CellTypeI As Long, CellTypeII As Long
Dim tmp_n_cells(-5 To 800) As Integer 'table x no. of contrib cells
Dim r(1 To 40) As Integer, c(1 To 40) As Integer

'Find dimensions of current table
Input #mapfile, rows(table_no), cols(table_no)

'print #temp, "Table "; t; " rows "; rows(t); " cols "; cols(t)

'Set all table cells in the array contrib_cells to 1

For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    contrib_cells(table_no, x, y) = 1
  Next x
Next y


mapping = 0

Do While Not EOF(mapfile)

   mapping = mapping + 1

  'Check to see if mapping is for row or column
  Input #mapfile, rowcol

  'Find no. of cols/rows wide table is, as appropriate

  If (rowcol = 1) Then
    map(mapping, 0) = 1
    x = cols(table_no) 'Mapping is for row X-cols wide
  End If
  If (rowcol = 2) Then
    map(mapping, 0) = 2
    x = rows(table_no) 'Mapping is for col X-rows wide
  End If

  'Read in mapping and identify marginal

  For y = 1 To x
    Input #mapfile, map(mapping, y)
    If (map(mapping, y) = -1) Then marginal = y
  Next y 'next row/col

  'print #temp, rowcol; " ";
  'For y = 1 to 30
    'print #temp, map(mapping, y); " ";
  'Next y
  'print #temp,
  
  'Count up number of cells contributing to marginal
  
  If (map(mapping, 0) = 1) Then 'col. marginal
    For y = 1 To rows(table_no)
      origvalue = contrib_cells(table_no, marginal, y)
      For x = 1 To cols(table_no)
        If map(mapping, x) > 0 Then contrib_cells(table_no, marginal, y) = _
           contrib_cells(table_no, marginal, y) + contrib_cells(table_no, x, y)
      Next x 'next col.
      contrib_cells(table_no, marginal, y) = contrib_cells(table_no, marginal, y) - origvalue
    Next y 'next row
  End If
  
  If (map(mapping, 0) = 2) Then 'row marginal
    For x = 1 To cols(table_no)
      origvalue = contrib_cells(table_no, x, marginal)
      For y = 1 To rows(table_no)
        If map(mapping, y) > 0 Then contrib_cells(table_no, x, marginal) = _
          contrib_cells(table_no, x, marginal) + contrib_cells(table_no, x, y)
      Next y 'next col.
      contrib_cells(table_no, x, marginal) = contrib_cells(table_no, x, marginal) - origvalue
    Next x 'next row
  End If

    'Input #original_data(T), table(ED, 0, 0, 1)

'  Print #temp, "Count_cells"; " Mapping "; mapping; " marginal ", marginal
'  For y = 1 to 30
'    For x = 1 to 30
'      Print #temp, f(contrib_cells(table_no,x, y), 5, 0);
'    Next x
'    Print #temp,
'  Next y
  
Loop 'next mapping for current table

NoOfMaps = mapping


'Identify cell containing table total (one w. most contributing cells)
MaxContribCells = 1
TableTotalX = 0 'No table marginal in input table
TableTotalY = 0 'No table marginal in input table
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    If (contrib_cells(table_no, x, y) > MaxContribCells) Then
      MaxContribCells = contrib_cells(table_no, x, y)
      TableTotalX = x
      TableTotalY = y
    End If
    If MaxContribCells > MaxContribCells_tot Then MaxContribCells_tot = MaxContribCells
  Next x
Next y
'Print #temp, "Total located in row "; y; " column "; x


'Report table mapping
If (report_table_mapping = 1) Then
  Print #temp, "=== Table mapping: "; Trim(tablename); " ==="
  Print #temp,
  Print #temp, "Total located in row "; TableTotalY; " column "; TableTotalX
  Print #temp, "Marginal vs. Internal cells"
  For y = 1 To rows(table_no)
    For x = 1 To cols(table_no)
      Print #temp, f(contrib_cells(table_no, x, y), 5, 0);
    Next x
    Print #temp,
  Next y
  Print #temp,
End If

'--- Find no. of rows and cols for table containing cells of given type

For celltype = 1 To MaxContribCells

  For x = 1 To cols(table_no)
    r(x) = 0
    For y = 1 To rows(table_no)
      If contrib_cells(table_no, x, y) = celltype Then _
        r(x) = r(x) + 1
    Next y
  Next x

  min_rows(table_no, celltype) = 0
  For x = 1 To cols(table_no)
    If r(x) > min_rows(table_no, celltype) Then min_rows(table_no, celltype) = r(x)
  Next x

  For y = 1 To rows(table_no)
    c(y) = 0
    For x = 1 To cols(table_no)
      If contrib_cells(table_no, x, y) = celltype Then _
        c(y) = c(y) + 1
    Next x
  Next y

  min_cols(table_no, celltype) = 0
  For y = 1 To rows(table_no)
    If c(y) > min_cols(table_no, celltype) Then min_cols(table_no, celltype) = c(y)
  Next y

Next celltype

min_cols(table_no, -1) = min_cols(table_no, 1)
min_cols(table_no, 0) = cols(table_no)
min_rows(table_no, -1) = min_rows(table_no, 1)
min_rows(table_no, 0) = rows(table_no)

'For marginal cells, there is no clear no. of rows & cols (c.f. SAS Table06a);
'therefore set to 1, ensuring Cramer's V is undefined.
min_cols(table_no, -2) = 1
min_rows(table_no, -2) = 1

'For each cell type, count no. of tables for which min_row and min_col are both >1
'[i.e. for which Cramer's V is not undefined]
For celltype = -2 To MaxContribCells
  If (min_cols(table_no, celltype) > 1) And (min_rows(table_no, celltype) > 1) Then
    n_tables_with_celltype(celltype) = _
      n_tables_with_celltype(celltype) + 1
  End If
Next celltype

'print #temp, "no. of mappings ", mapping


For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)

'Identify Cell Type I [internal or marginal]
If (contrib_cells(table_no, x, y) = 1) Then CellTypeI = -1 Else CellTypeI = -2
    
'Identify Cell Type II [based on number of cells contributing to error]
'(value ranges from 1 to z, where z is max. no. of cells contributing to
'calculation of a marginal count)
CellTypeII = contrib_cells(table_no, x, y)
    
'Cumulate number of cells of each type
tmp_n_cells(CellTypeI) = tmp_n_cells(CellTypeI) + 1 'Internal/Marginal cells
tmp_n_cells(CellTypeII) = tmp_n_cells(CellTypeII) + 1 'No. of contrib. cells
tmp_n_cells(0) = tmp_n_cells(0) + 1 'All table cells


Next x
Next y


' ========= For current table, identify frequency of cells of different types
'           [where 'cell type' = no. of cell counts contributing to given cell value]

'Identify no. of valid Cell types
'[where 'cell type' = no. of cells contributing to cell count]

n_valid_Celltypes = 0
celltype_map(-2) = -2
celltype_map(-1) = -1
celltype_map(0) = 0

For celltype = -2 To 0
  n_cells(celltype) = tmp_n_cells(celltype)
  n_cells_tot(celltype) = n_cells_tot(celltype) + n_cells(celltype)
Next celltype

For celltype = 1 To MaxContribCells
  If (tmp_n_cells(celltype) > 0) Then
    n_valid_Celltypes = n_valid_Celltypes + 1
    celltype_map(n_valid_Celltypes) = celltype
    inv_celltype_map(celltype) = n_valid_Celltypes
    n_cells(n_valid_Celltypes) = tmp_n_cells(celltype)
    n_cells_tot(celltype) = n_cells_tot(celltype) + tmp_n_cells(celltype)
  End If
Next celltype


'--- REPORT TABLE MAPPING (if required) ---

If (report_table_mapping = 1) Then
  
  Print #temp, "No. of contributing cells (Cell Type)", "No. of cells of this type", _
   "Min cols", "Min rows"
  Print #temp, "Marginal ", n_cells(-2), min_cols(table_no, -2), min_rows(table_no, -2)
  Print #temp, "Internal ", n_cells(1), min_cols(table_no, -1), min_rows(table_no, -1)
  Print #temp, "All ", n_cells(0), min_cols(table_no, 0), min_rows(table_no, 0)
  
  For x = 1 To n_valid_Celltypes
    Print #temp, celltype_map(x), n_cells(x), _
      min_cols(table_no, celltype_map(x)), min_rows(table_no, celltype_map(x))
      'n_cells(celltype_map(x))
  Next x
  
  Print #temp,

End If

End Sub


Sub get_sample_tables(table_no, sample, _
  TableTotalX, TableTotalY, original_data, perturbed_data, temp)

Dim s As Integer, x As Integer, y As Integer, ED As Long
Dim mapping As Integer, marginal As Integer, total1 As Double, total2 As Double
Dim temp_table(1 To 20, 1 To 40, 1 To 2) As Double
Dim exp_header As String, obs_header As String
Dim c As Integer, no_of_char As Integer, stop_checking As Integer
Dim line_end_blank As Integer

'Read in and check header only if a count table
If Count_table_in_use = 1 Then

  '(1) Read in data from current Table for current sample
      
  Input #original_data, exp_header
  Input #perturbed_data, obs_header
    
  '(2) Check that exp_header = obs_header (except for variant no.)
      
  '(i) find length of header (in letters)
  no_of_char = Len(exp_header)
      
  '(ii) check for similarity character by character, excpect for
  '     variant number found between v and next underscore
      
  stop_checking = 0
  For c = 1 To no_of_char
    If (Mid(exp_header, c, 1) = "v") Then stop_checking = 1
    If ((Mid(exp_header, c, 1) = "_") And (stop_checking = 1)) Then _
      stop_checking = 0
    If (stop_checking = 0) Then 'only check if not variant no.
      If (Mid(exp_header, c, 1) <> Mid(obs_header, c, 1)) Then
        MsgBox ("ERROR: sample headers do not match")
        Stop
      End If
    End If
  Next c
    
End If 'Count_table_in_use = 1

'expected (original) count = cell_exp(sample,x,y,1);
'observed (perturbed) count = cell_obs(sample,x,y,2)
  For y = 1 To rows(table_no)
    For x = 1 To cols(table_no)
      Input #original_data, cell_exp(table_no, sample, x, y)
      Input #perturbed_data, cell_obs(table_no, sample, x, y)
    Next x
    If Count_table_in_use = 0 Then
      Input #original_data, line_end_blank
      Input #perturbed_data, line_end_blank
    End If
  Next y

'Find total of internal counts for cumulated sample table
'(not necessarily same as table total as recorded in perturbed table due to perturbation!)
'To do so requires the creation of a temporary version of the cumulated sample table
'for calculation and storage of the table marginals consistent with table internal counts

'Make temporary copy of cumulated sample table
For x = 1 To cols(table_no)
  For y = 1 To rows(table_no)
    temp_table(x, y, 1) = cell_exp(table_no, sample, x, y)
    temp_table(x, y, 2) = cell_obs(table_no, sample, x, y)
  Next y
Next x

For mapping = 1 To NoOfMaps

If (map(mapping, 0) = 2) Then 'row marginal
  For x = 1 To cols(table_no)
    total2 = 0
    total1 = 0
    marginal = 0
    For y = 1 To rows(table_no)
      If (map(mapping, y) < 0) Then marginal = y
      If (map(mapping, y) > 0) Then total2 = total2 + temp_table(x, y, 2)
      If (map(mapping, y) > 0) Then total1 = total1 + temp_table(x, y, 1)
    Next y 'next row
    temp_table(x, marginal, 2) = total2
  Next x 'next col
End If 'if row marginal

If (map(mapping, 0) = 1) Then 'col marginal
  For y = 1 To rows(table_no)
    total2 = 0
    total1 = 0
    marginal = 0
    For x = 1 To cols(table_no)
      If (map(mapping, x) < 0) Then marginal = x
      If (map(mapping, x) > 0) Then total2 = total2 + temp_table(x, y, 2)
      If (map(mapping, x) > 0) Then total1 = total1 + temp_table(x, y, 1)
    Next x 'next col
    temp_table(marginal, y, 2) = total2
    temp_table(marginal, y, 1) = total1
  Next y 'next col
End If 'if row marginal

Next mapping

If ((TableTotalX = 0) Or (TableTotalY = 0)) Then
  'Need to calculate obs/exp table totals because supplied table consists
  'of internal table counts only
  obs_table_total = 0 'MUST be reset to 0 for each new sample!
  exp_table_total = 0 'MUST be reset to 0 for each new sample!
  For x = 1 To cols(table_no)
    For y = 1 To rows(table_no)
      obs_table_total = obs_table_total + temp_table(x, y, 2)
      exp_table_total = exp_table_total + temp_table(x, y, 1)
    Next y
  Next x
Else
  obs_table_total = temp_table(TableTotalX, TableTotalY, 2)
  exp_table_total = temp_table(TableTotalX, TableTotalY, 1)
End If


If (report_table_totals = 1) Then
  Print #temp, "=== Revised table totals for "; tablename; " ==="
  Print #temp,
    
  If ((TableTotalX = 0) Or (TableTotalY = 0)) Then
    Print #temp, "Table "; tablename; " Supplied table contains no table total - internal counts only"
  Else
    Print #temp, "Table "; tablename; " As published          : Expected total "; _
      cell_exp(table_no, sample, TableTotalX, TableTotalY); " Observed total "; _
      cell_obs(table_no, sample, TableTotalX, TableTotalY)
  End If
  
  Print #temp, "Table "; tablename; " Sum of internal counts: Expected total "; exp_table_total; _
                          " Observed total "; obs_table_total
  Print #temp,
End If

End Sub
Sub Measure_diffs(table_no, TableTotalX, TableTotalY, _
     contrib_cells, temp, sample)
  
'Temporary internal variables
Dim version As Integer, x As Integer, y As Integer, minRC As Integer
Dim celltype As Integer, CellTypeO As Integer, CellTypeI As Integer, CellTypeII As Integer
Dim p_change As Double
Dim observed As Double, expected As Double
Dim ave_table_total As Double

Dim tmp_obs(1 To 800) As Long, tmp_exp(1 To 800) As Long 'table as vector (1:max poss. cells in table)

Dim obsVector(-2 To 50, 1 To 800) As Double 'obs value by contributing cells (mapped); no. of cells of this type in table
Dim expVector(-2 To 50, 1 To 800) As Double 'exp value by contributing cells (mapped); no. of cells of this type in table

Dim nCellType(-2 To 50) As Integer 'no. of contributing cells by cell type (mapped)
Dim absdiff(1 To 20, 1 To 40) As Double

'(0) Initialise subroutine specific arrrays that assume starting value of zero

For CellTypeI = -2 To 50
  nCellType(CellTypeI) = 0
Next CellTypeI

'(1) Caculate all cell-level measures required to assess impact of disclosure control

For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    
    'Note observed (perturbed) and expected (original) cell count
    expected = cell_exp(table_no, sample, x, y)
    observed = cell_obs(table_no, sample, x, y)
      
    '======================================================================
    '======================  1. CELL-BASED MEASURES =======================
    '======================================================================
    
    'Find cell error
    cell_diff(sample, x, y) = observed - expected
    
    'Find absolute cell error
    absdiff(x, y) = Abs(observed - expected)
    
    'Find cell Z-score
    cell_Z(sample, x, y) = f_Z(observed, expected, obs_table_total, exp_table_total)
      
    'Find cell Zm-score
    cell_Zm(sample, x, y) = f_Zm(observed, expected, obs_table_total, exp_table_total)
      
    'Find change as % of original cell value
    If (expected <> 0) Then '% change poss. only if orginal value<>0
      p_change = Abs(cell_diff(sample, x, y)) / cell_exp(table_no, sample, x, y) * 100#
    Else
      p_change = 0#
    End If
      
      
      
    '===================== Identify cell type ========================
      
    'Actual No. of contributing cells [shorthand]
    CellTypeO = contrib_cells(table_no, x, y)
      
    'Identify Cell Type I [internal or marginal]
    If (contrib_cells(table_no, x, y) = 1) Then CellTypeI = -1 Else CellTypeI = -2
    
    'Identify Cell Type II [based on number of cells contributing to error]
    '(value ranges from 1 to z, where z is max. no. of cells contributing to
    'calculation of a marginal count);
    '[To save space/time uses mapping of actual cell type onto mapped celltype]
    
'    CellTypeII = contrib_cells(T, x, y)
    CellTypeII = inv_celltype_map(contrib_cells(table_no, x, y))

    '======================================================================
    '====================== 2. TABLE-BASED MEASURES =======================
    '====================== (by type of cell) =============================
          
    'Cumulate original values for each type of cell
    '[needed to calculate SAE and Cramers V]
    v_expCells(sample, CellTypeI) = v_expCells(sample, CellTypeI) + expected 'Internal/Marginal cells
    v_expCells(sample, CellTypeII) = v_expCells(sample, CellTypeII) + expected 'No. of contrib. cells
    v_expCells(sample, 0) = v_expCells(sample, 0) + expected 'All table cells
    v_expCells_tot(sample, CellTypeI) = v_expCells_tot(sample, CellTypeI) + expected 'Internal/Marginal cells
    v_expCells_tot(sample, CellTypeO) = v_expCells_tot(sample, CellTypeO) + expected 'No. of contrib. cells
    v_expCells_tot(sample, 0) = v_expCells_tot(sample, 0) + expected 'All table cells

    v_obsCells(sample, CellTypeI) = v_obsCells(sample, CellTypeI) + observed 'Internal/Marginal cells
    v_obsCells(sample, CellTypeII) = v_obsCells(sample, CellTypeII) + observed 'No. of contrib. cells
    v_obsCells(sample, 0) = v_obsCells(sample, 0) + observed 'All table cells
    v_obsCells_tot(sample, CellTypeI) = v_obsCells_tot(sample, CellTypeI) + observed 'Internal/marginal cells
    v_obsCells_tot(sample, CellTypeO) = v_obsCells_tot(sample, CellTypeO) + observed 'No. of contrib. cells
    v_obsCells_tot(sample, 0) = v_obsCells_tot(sample, 0) + observed 'All table cells

    If (Abs(cell_Z(sample, x, y)) > 1.96) Then
      cell_NFC(sample, x, y) = 1
      NFC(sample, CellTypeI) = NFC(sample, CellTypeI) + 1   'Internal/marginal cells
      NFC(sample, CellTypeII) = NFC(sample, CellTypeII) + 1   'No. of contrib. cells
      NFC(sample, 0) = NFC(sample, 0) + 1   'All table cells
      NFC_tot(sample, CellTypeI) = NFC_tot(sample, CellTypeI) + 1   'Internal/marginal cells
      NFC_tot(sample, CellTypeO) = NFC_tot(sample, CellTypeO) + 1   'No. of contrib. cells
      NFC_tot(sample, 0) = NFC_tot(sample, 0) + 1   'All table cells
    End If

    If (Abs(cell_Zm(sample, x, y)) > 1.96) Then
      cell_NFCm(sample, x, y) = 1
      NFCm(sample, CellTypeI) = NFCm(sample, CellTypeI) + 1   'Internal/marginal cells
      NFCm(sample, CellTypeII) = NFCm(sample, CellTypeII) + 1   'No. of contrib. cells
      NFCm(sample, 0) = NFCm(sample, 0) + 1   'All table cells
      NFCm_tot(sample, CellTypeI) = NFCm_tot(sample, CellTypeI) + 1   'Internal/marginal cells
      NFCm_tot(sample, CellTypeO) = NFCm_tot(sample, CellTypeO) + 1   'No. of contrib. cells
      NFCm_tot(sample, 0) = NFCm_tot(sample, 0) + 1   'All table cells
    End If

    If (cell_diff(sample, x, y) <> 0) Then  'If error<>0
    
      'Note that cell has been changed
      cell_changed(sample, x, y) = 1
    
      'Cumulate number of changed cells
      n_changed(sample, CellTypeI) = n_changed(sample, CellTypeI) + 1   'Internal/Marginal cells
      n_changed(sample, CellTypeII) = n_changed(sample, CellTypeII) + 1   'No. of contrib. cells
      n_changed(sample, 0) = n_changed(sample, 0) + 1   'All table cells
      n_changed_tot(sample, CellTypeI) = n_changed_tot(sample, CellTypeI) + 1
      n_changed_tot(sample, CellTypeO) = n_changed_tot(sample, CellTypeO) + 1
      n_changed_tot(sample, 0) = n_changed_tot(sample, 0) + 1   'All table cells
      
      'Cumulate Total Error
      TE(sample, CellTypeI) = TE(sample, CellTypeI) + cell_diff(sample, x, y)    'Internal/Marginal cells
      TE(sample, CellTypeII) = TE(sample, CellTypeII) + cell_diff(sample, x, y)   'No. of contrib. cells
      TE(sample, 0) = TE(sample, 0) + cell_diff(sample, x, y)    'All table cells
      TE_tot(sample, CellTypeI) = TE_tot(sample, CellTypeI) + cell_diff(sample, x, y)    'Internal/Marginal cells
      TE_tot(sample, CellTypeO) = TE_tot(sample, CellTypeO) + cell_diff(sample, x, y)   'No. of contrib. cells
      TE_tot(sample, 0) = TE_tot(sample, 0) + cell_diff(sample, x, y)    'All table cells
      
      'Cumulate Total Absolute Error
      TAE(sample, CellTypeI) = TAE(sample, CellTypeI) + Abs(cell_diff(sample, x, y))   'Internal/Marginal cells
      TAE(sample, CellTypeII) = TAE(sample, CellTypeII) + Abs(cell_diff(sample, x, y))   'No. of contrib. cells
      TAE(sample, 0) = TAE(sample, 0) + Abs(cell_diff(sample, x, y))    'All table cells
      TAE_tot(sample, CellTypeI) = TAE_tot(sample, CellTypeI) + Abs(cell_diff(sample, x, y))   'Internal/Marginal cells
      TAE_tot(sample, CellTypeO) = TAE_tot(sample, CellTypeO) + Abs(cell_diff(sample, x, y))   'No. of contrib. cells
      TAE_tot(sample, 0) = TAE_tot(sample, 0) + Abs(cell_diff(sample, x, y))    'All table cells

      'Cumulate Square error
      SE(sample, CellTypeI) = SE(sample, CellTypeI) + cell_diff(sample, x, y) ^ 2    'Internal/marginal cells
      SE(sample, CellTypeII) = SE(sample, CellTypeII) + cell_diff(sample, x, y) ^ 2    'No. of contrib cells
      SE(sample, 0) = SE(sample, 0) + cell_diff(sample, x, y) ^ 2    'All table cells
      SE_tot(sample, CellTypeI) = SE_tot(sample, CellTypeI) + cell_diff(sample, x, y) ^ 2    'Internal/marginal cells
      SE_tot(sample, CellTypeO) = SE_tot(sample, CellTypeO) + cell_diff(sample, x, y) ^ 2    'No. of contrib cells
      SE_tot(sample, 0) = SE_tot(sample, 0) + cell_diff(sample, x, y) ^ 2    'All table cells
      
      'Cumulate Total Values of Changed Counts
      TVCC(sample, CellTypeI) = TVCC(sample, CellTypeI) + expected   'Internal/Marginal cells
      TVCC(sample, CellTypeII) = TVCC(sample, CellTypeII) + expected   'No. of contrib. cells
      TVCC(sample, 0) = TVCC(sample, 0) + expected   'All table cells
      TVCC_tot(sample, CellTypeI) = TVCC_tot(sample, CellTypeI) + expected
      TVCC_tot(sample, CellTypeO) = TVCC_tot(sample, CellTypeO) + expected
      TVCC_tot(sample, 0) = TVCC_tot(sample, 0) + expected

    End If
    
    'Note maximum cell change in table
    If Abs(cell_diff(sample, x, y)) > max_change(sample, CellTypeI) Then _
       max_change(sample, CellTypeI) = Abs(cell_diff(sample, x, y))
    If Abs(cell_diff(sample, x, y)) > max_change(sample, CellTypeII) Then _
       max_change(sample, CellTypeII) = Abs(cell_diff(sample, x, y))
    If Abs(cell_diff(sample, x, y)) > max_change(sample, 0) Then _
       max_change(sample, 0) = Abs(cell_diff(sample, x, y))
    
    'Note maximum cell change across all tables
    If max_change(sample, CellTypeI) > max_change_tot(sample, CellTypeI) Then _
      max_change_tot(sample, CellTypeI) = max_change(sample, CellTypeI)
    If max_change(sample, 0) > max_change_tot(sample, 0) Then _
      max_change_tot(sample, 0) = max_change(sample, 0)
    If max_change(sample, CellTypeII) > max_change_tot(sample, CellTypeO) Then _
      max_change_tot(sample, CellTypeO) = max_change(sample, CellTypeII)
      
    'Note maximum % cell change in table
    If p_change > max_p_change(sample, CellTypeI) Then _
       max_p_change(sample, CellTypeI) = p_change
    If p_change > max_p_change(sample, CellTypeII) Then _
       max_p_change(sample, CellTypeII) = p_change
    If p_change > max_p_change(sample, 0) Then _
       max_p_change(sample, 0) = p_change
        
    'Note maximum % cell change across all tables
    If p_change > max_p_change_tot(sample, CellTypeI) Then _
      max_p_change_tot(sample, CellTypeI) = p_change
    If p_change > max_p_change_tot(sample, 0) Then _
      max_p_change_tot(sample, 0) = p_change
    If p_change > max_p_change_tot(sample, CellTypeO) Then _
      max_p_change_tot(sample, CellTypeO) = p_change
        
    'Cumulate Sum of Squared Z-scores
    SSZ(sample, CellTypeI) = SSZ(sample, CellTypeI) + cell_Z(sample, x, y) ^ 2    'Internal/Marginal SSZ
    SSZ(sample, CellTypeII) = SSZ(sample, CellTypeII) + cell_Z(sample, x, y) ^ 2    'No. of contrib. cells SSZ
    SSZ(sample, 0) = SSZ(sample, 0) + cell_Z(sample, x, y) ^ 2    'Overall table SSZ
    SSZ_tot(sample, CellTypeI) = SSZ_tot(sample, CellTypeI) + cell_Z(sample, x, y) ^ 2    'Internal/Marginal SSZ
    SSZ_tot(sample, CellTypeO) = SSZ_tot(sample, CellTypeO) + cell_Z(sample, x, y) ^ 2    'No. of contrib. cells SSZ
    SSZ_tot(sample, 0) = SSZ_tot(sample, 0) + cell_Z(sample, x, y) ^ 2    'Overall table SSZ
    
    'Cumulate Sum of Squared Zm-scores
    SSZm(sample, CellTypeI) = SSZm(sample, CellTypeI) + cell_Zm(sample, x, y) ^ 2    'Internal/Marginal SSZ
    SSZm(sample, CellTypeII) = SSZm(sample, CellTypeII) + cell_Zm(sample, x, y) ^ 2   'No. of contrib. cells SSZ
    SSZm(sample, 0) = SSZm(sample, 0) + cell_Zm(sample, x, y) ^ 2   'Overall table SSZ
    SSZm_tot(sample, CellTypeI) = SSZm_tot(sample, CellTypeI) + cell_Zm(sample, x, y) ^ 2    'Internal/Marginal SSZ
    SSZm_tot(sample, CellTypeO) = SSZm_tot(sample, CellTypeO) + cell_Zm(sample, x, y) ^ 2   'No. of contrib. cells SSZ
    SSZm_tot(sample, 0) = SSZm_tot(sample, 0) + cell_Zm(sample, x, y) ^ 2   'Overall table SSZ
    
    'Cumulate chisquare
    If (expected > 0) Then '[If expected = 0, assume observed=0; therefore no addition to chi-square]
      ChiSquare(sample, CellTypeI) = ChiSquare(sample, CellTypeI) + (observed - expected) ^ 2 / expected
      ChiSquare(sample, CellTypeII) = ChiSquare(sample, CellTypeII) + (observed - expected) ^ 2 / expected
      ChiSquare(sample, 0) = ChiSquare(sample, 0) + (observed - expected) ^ 2 / expected
      ChiSquare_tot(sample, CellTypeI) = ChiSquare_tot(sample, CellTypeI) + (observed - expected) ^ 2 / expected
      ChiSquare_tot(sample, CellTypeO) = ChiSquare_tot(sample, CellTypeO) + (observed - expected) ^ 2 / expected
      ChiSquare_tot(sample, 0) = ChiSquare_tot(sample, 0) + (observed - expected) ^ 2 / expected
    End If
       
    'Create vectorised versions of array for each cell type
    '[first augmenting count of cells of relevant type for current cell by +1]
    nCellType(CellTypeI) = nCellType(CellTypeI) + 1
    nCellType(CellTypeII) = nCellType(CellTypeII) + 1
    nCellType(0) = nCellType(0) + 1
    obsVector(CellTypeI, nCellType(CellTypeI)) = observed
    obsVector(CellTypeII, nCellType(CellTypeII)) = observed
    obsVector(0, nCellType(0)) = observed
    expVector(CellTypeI, nCellType(CellTypeI)) = expected
    expVector(CellTypeII, nCellType(CellTypeII)) = expected
    expVector(0, nCellType(0)) = expected
           
  Next x
Next y

'Calculate table based measures
'[if not already calculated by cumulation of cell-based measures]

'Find % of table cells perturbed [set to 0 if no cells of given type exist]
For celltype = -2 To n_valid_Celltypes
'For x = -2 To n_valid_Celltypes
  'CellType = celltype_map(x)
  If (n_cells(celltype) > 0) Then
    p_changed(sample, celltype) = n_changed(sample, celltype) / n_cells(celltype) * 100#
  Else
    p_changed(sample, celltype) = 0#
  End If
Next celltype
'Next x

'Caclulate RAE [TAE as % of total original values of changed cells]
For celltype = -2 To n_valid_Celltypes
'For x = -2 To n_valid_Celltypes
'  CellType = celltype_map(x)
  If (TVCC(sample, celltype) > 0) Then
    RAE(sample, celltype) = TAE(sample, celltype) / TVCC(sample, celltype) * 100#
  Else
    RAE(sample, celltype) = 0#
  End If
Next celltype
'Next x

'Calculate SAE [=TAE/total of original [expected] cell values]
For celltype = -2 To n_valid_Celltypes
'For x = -2 To n_valid_Celltypes
'  CellType = celltype_map(x)
  If (v_expCells(sample, celltype) > 0) Then
    SAE(sample, celltype) = TAE(sample, celltype) / v_expCells(sample, celltype)
  Else
    SAE(sample, celltype) = 0#
  End If
Next celltype
'Next x

'Calculate NFT (whether or not table SSZ <= chi-square at p=0.05, df=no. of table cells)
For celltype = -2 To n_valid_Celltypes
  If (SSZ(sample, celltype) > chiinv(n_cells(celltype))) Then
    NFT(sample, celltype) = 1
    NFT_tot(sample, celltype_map(celltype)) = _
      NFT_tot(sample, celltype_map(celltype)) + 1
  Else
    NFT(sample, celltype) = 0
  End If
Next celltype


'Calculate NFT (whether or not table SSZ <= chi-square at p=0.05, df=no. of table cells)
For celltype = -2 To n_valid_Celltypes
'For x = -2 To n_valid_Celltypes
'  CellType = celltype_map(x)
  If (SSZm(sample, celltype) > chiinv(n_cells(celltype))) Then
    NFTm(sample, celltype) = 1
    NFTm_tot(sample, celltype_map(celltype)) = _
      NFTm_tot(sample, celltype_map(celltype)) + 1
  Else
    NFTm(sample, celltype) = 0
  End If
Next celltype
'Next x


'Calculate RMSE (=square root of average sq_error for given cell type)
For celltype = -2 To n_valid_Celltypes
'For x = -2 To n_valid_Celltypes
'  CellType = celltype_map(x)
  If (n_cells(celltype) > 0) Then
    RMSE(sample, celltype) = (SE(sample, celltype) / n_cells(celltype)) ^ 0.5
  Else
    RMSE(sample, celltype) = 0#
  End If
Next celltype
'Next x

'Calculate CramersV for each Cell Type
For celltype = -2 To n_valid_Celltypes
  minRC = min_rows(table_no, celltype_map(celltype))
  If min_cols(table_no, celltype_map(celltype)) < minRC Then _
    minRC = min_cols(table_no, celltype_map(celltype))
  ave_table_total = (v_expCells(sample, celltype) + v_obsCells(sample, celltype)) / 2
  'If ((ave_table_total > 0) And (n_cells(celltype) > 1)) Then
  If ((ave_table_total > 0) And (minRC > 1)) Then
    CramersV(sample, celltype) = _
      (ChiSquare(sample, celltype) / (ave_table_total * (minRC - 1))) ^ 0.5
    CramersV_tot(sample, celltype_map(celltype)) = _
      CramersV_tot(sample, celltype_map(celltype)) + CramersV(sample, celltype)
  Else
    CramersV(sample, celltype) = -9  'If ave_table_total = 0 or minRC = 1
  End If
Next celltype

' To calculate Gibsons D and Pearsons R, first need to create cell-type specific
' vector arrays of original and observed cell counts

For celltype = -2 To n_valid_Celltypes
'For y = -2 To n_valid_Celltypes

'    CellType = celltype_map(y)

    'Set up temporary vectors
    For x = 1 To nCellType(celltype)
      tmp_obs(x) = obsVector(celltype, x)
      tmp_exp(x) = expVector(celltype, x)
    Next x

    'Cacluate Gibsons D for current Cell Type
    D(sample, celltype) = f_GibsonsD(tmp_obs, tmp_exp, nCellType(celltype))

 If nCellType(celltype) > 1 Then  'Can only calculate R for 2+ cells
    
    'Caclulate PearsonsR for current Cell Type
    PearsonsR(sample, celltype) = f_PearsonsR(tmp_obs, tmp_exp, nCellType(celltype))
    
  Else 'statistic undefined because < 2 cells of current cell type
    
    PearsonsR(sample, celltype) = -9

  End If
  
Next celltype
'Next y

  'Write out summary stats for current table (if required)

  'a) Cell-level

  If report_cell_diffs_flag = 1 Then _
    Call report_cell_diffs(sample, table_no, temp)

  'b) Table-level

  If report_table_diffs_flag = 1 Then _
    Call report_table_diffs(sample, temp)

End Sub
Sub Calc_Crosstable_Diffs(contrib_cells)

Dim celltype As Integer, sample As Integer
Dim CellTypeII As Integer, x As Integer, y As Integer, cell_no As Integer
Dim obsVector(1 To 16000) As Double, expVector(1 To 16000) As Double
Dim table_no As Integer

For sample = 1 To NoOfSamples

'Calculate table based measures
'[if not already calculated by cumulation of cell-based measures]

'Find % of table cells perturbed [set to 0 if no cells of given type exist]
For celltype = -2 To MaxContribCells_tot
  If (n_cells_tot(celltype) > 0) Then
    p_changed_tot(sample, celltype) = n_changed_tot(sample, celltype) / n_cells_tot(celltype) * 100#
  Else
    p_changed_tot(sample, celltype) = 0#
  End If
Next celltype

'Caclulate RAE [TAE as % of total original values of changed cells]
For celltype = -2 To MaxContribCells_tot
  If (TVCC_tot(sample, celltype) > 0) Then
    RAE_tot(sample, celltype) = TAE_tot(sample, celltype) / TVCC_tot(sample, celltype) * 100#
  Else
    RAE_tot(sample, celltype) = 0#
  End If
Next celltype

'Calculate SAE [=TAE/total of original [expected] cell values]
For celltype = -2 To MaxContribCells_tot
  If (v_expCells_tot(sample, celltype) > 0) Then
    SAE_tot(sample, celltype) = TAE_tot(sample, celltype) / v_expCells_tot(sample, celltype)
  Else
    SAE_tot(sample, celltype) = 0#
  End If
Next celltype


'Calculate RMSE (=square root of average sq_error for given cell type)
For celltype = -2 To MaxContribCells_tot
  If (n_cells_tot(celltype) > 0) Then
    RMSE_tot(sample, celltype) = (SE_tot(sample, celltype) / n_cells_tot(celltype)) ^ 0.5
  Else
    RMSE_tot(sample, celltype) = 0#
  End If
Next celltype


'Calculate cross-table CramersV for each Cell Type
'[= average of summed table-specific CramersV, as min of no. of rows and no. of cols
'has no interpretable cross-table meaning; averaged only across tables for which
'min row/col >1 for given celltype]

For celltype = -2 To MaxContribCells_tot
  
  If n_tables_with_celltype(celltype) > 0 Then
    CramersV_tot(sample, celltype) = CramersV_tot(sample, celltype) / _
      n_tables_with_celltype(celltype)
  Else
    CramersV_tot(sample, celltype) = -9
  End If
 
Next celltype

' To calculate Gibsons D and Pearsons R, first need to create cell-type specific
' vector array of original and observed cell counts

For celltype = -2 To MaxContribCells_tot

  cell_no = 0

  If n_cells_tot(celltype) > 0 Then

    For table_no = 1 To NoOfTables

      For x = 1 To cols(table_no)

        For y = 1 To rows(table_no)
          
          If celltype < 0 Then
            If (contrib_cells(table_no, x, y) = 1) Then
              CellTypeII = -1 'Internal
            Else
              CellTypeII = -2 'Marginal
            End If
          ElseIf celltype = 0 Then
            CellTypeII = 0
          Else
            CellTypeII = contrib_cells(table_no, x, y)
          End If
          
          If (celltype = CellTypeII) Then
            cell_no = cell_no + 1
            obsVector(cell_no) = cell_exp(table_no, sample, x, y)
            expVector(cell_no) = cell_obs(table_no, sample, x, y)
          End If
          
        Next y
    
      Next x
  
    Next table_no
  
    D_tot(sample, celltype) = f_GibsonsD(obsVector, expVector, cell_no)
    
    If cell_no > 1 Then 'can only calculate R for 2+ cells
      PearsonsR_tot(sample, celltype) = f_PearsonsR(obsVector, expVector, cell_no)
    Else 'statistic undefined because < 2 cells of current cell type
      PearsonsR_tot(sample, celltype) = -9
    End If
    
  Else 'no cells of current type
    
    D_tot(sample, celltype) = 0#
    PearsonsR_tot(sample, celltype) = 0#
      
  End If
  
Next celltype

Next sample


End Sub

Sub report_cell_diffs(sample, table_no, temp)
Dim x As Integer, y As Integer, fw As Integer, dp As Integer

If Count_table_in_use = 1 Then
  fw = 8
  dp = 0
Else
  fw = 10
  dp = 4
End If

'a) Cell-level

Print #temp, "=== Table-specific, Area-specific, Cell-based report for "; tablename; _
  " (Sample "; f(sample, 5, 0); ") ==="
Print #temp,

If (Cell_Fitness_Switch(1) = 1) Then
Print #temp, "cell_exp";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_exp(table_no, sample, x, y), fw, dp);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(2) = 1) Then
Print #temp, "cell_obs";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
      Print #temp, f(cell_obs(table_no, sample, x, y), fw, dp);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(3) = 1) Then
Print #temp, "cell_changed";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_changed(sample, x, y), fw, 0);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(4) = 1) Then
Print #temp, "cell_diff";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_diff(sample, x, y), fw, dp);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(5) = 1) Then
Print #temp, "cell_Z";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_Z(sample, x, y), fw, 4);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(6) = 1) Then
Print #temp, "cell_NFC";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_NFC(sample, x, y), fw, 0);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(7) = 1) Then
Print #temp, "cell_Zm";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_Zm(sample, x, y), fw, 4);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If (Cell_Fitness_Switch(8) = 1) Then
Print #temp, "cell_NFCm";
If Count_table_in_use = 1 Then Print #temp,
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
    Print #temp, f(cell_NFCm(sample, x, y), fw, 0);
  Next x
  Print #temp,
Next y
If rows(table_no) > 1 Then Print #temp, 'blank line between measures
End If

If rows(table_no) = 1 Then Print #temp, 'blank line after final measure

End Sub
Sub report_table_diffs(sample, temp)
Dim celltype As Long

'Check whether or not header is needed; write out if it is
If (report_table_diffs_header_flag = 1) Then
  Print #temp, "=== Table-specific, Area-specific, Table-based report for "; tablename;
  If (report_table_totals = 1 Or report_cell_diffs_flag = 1) Then Print #temp, " (Sample"; f(sample, 6, 0); ") ";
  Print #temp, " ==="
  Print #temp,

  Print #temp, "                    Cell type (no. of contributing cells count depends upon)"
  Print #temp, "Sample  Measure       Marginal    Internal    All";
  For celltype = 1 To n_valid_Celltypes
    Print #temp, f(celltype_map(celltype), 12, 0);
  Next celltype
  Print #temp,
End If

'After writing header out once, switch off table header flag - unless
'reporting table totals between table diffs
If (report_table_totals = 0 And report_cell_diffs_flag = 0 And _
  report_table_diffs_header_flag = 1) Then
  report_table_diffs_header_flag = 0
End If

If (Table_Fitness_Switch(0) = 1) Then
Print #temp, f(sample, 6, 0); "  frequency";
For celltype = -2 To n_valid_Celltypes
  If celltype = -2 Then
    Print #temp, f(n_cells(celltype), 8, 0);
  Else
    Print #temp, f(n_cells(celltype), 12, 0);
  End If
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(1) = 1) Then
Print #temp, f(sample, 6, 0); "  n_changed   ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(n_changed(sample, celltype), 12, t_dp(1));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(2) = 1) Then
Print #temp, f(sample, 6, 0); "  p_changed   ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(p_changed(sample, celltype), 12, t_dp(2));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(3) = 1) Then
Print #temp, f(sample, 6, 0); "  max_change  ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(max_change(sample, celltype), 12, t_dp(3));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(4) = 1) Then
Print #temp, f(sample, 6, 0); "  max_p_change";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(max_p_change(sample, celltype), 12, t_dp(4));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(5) = 1) Then
Print #temp, f(sample, 6, 0); "  TotalError  ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(TE(sample, celltype), 12, t_dp(5));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(6) = 1) Then
Print #temp, f(sample, 6, 0); "  TAE         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(TAE(sample, celltype), 12, t_dp(6));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(7) = 1) Then
Print #temp, f(sample, 6, 0); "  RAE         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(RAE(sample, celltype), 12, t_dp(7));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(21) = 1) Then
Print #temp, f(sample, 6, 0); "  TVCC        ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(TVCC(sample, celltype), 12, 2);
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(8) = 1) Then
Print #temp, f(sample, 6, 0); "  SAE         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(SAE(sample, celltype), 12, t_dp(8));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(22) = 1) Then
Print #temp, f(sample, 6, 0); "  v_expcells  ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(v_expCells(sample, celltype), 12, 2);
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(9) = 1) Then
Print #temp, f(sample, 6, 0); "  Sq_Error    ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(SE(sample, celltype), 12, t_dp(9));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(10) = 1) Then
Print #temp, f(sample, 6, 0); "  RMSE        ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(RMSE(sample, celltype), 12, t_dp(10));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(11) = 1) Then
Print #temp, f(sample, 6, 0); "  SSZ         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(SSZ(sample, celltype), 12, t_dp(11));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(12) = 1) Then
Print #temp, f(sample, 6, 0); "  NFC         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(NFC(sample, celltype), 12, t_dp(12));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(13) = 1) Then
Print #temp, f(sample, 6, 0); "  NFT         ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(NFT(sample, celltype), 12, t_dp(13));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(14) = 1) Then
Print #temp, f(sample, 6, 0); "  SSZm        ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(SSZm(sample, celltype), 12, t_dp(14));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(15) = 1) Then
Print #temp, f(sample, 6, 0); "  NFCm        ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(NFCm(sample, celltype), 12, t_dp(15));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(16) = 1) Then
Print #temp, f(sample, 6, 0); "  NFTm        ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(NFTm(sample, celltype), 12, t_dp(16));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(17) = 1) Then
Print #temp, f(sample, 6, 0); "  D           ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(D(sample, celltype), 12, t_dp(17));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(18) = 1) Then
Print #temp, f(sample, 6, 0); "  CramersV    ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(CramersV(sample, celltype), 12, t_dp(18));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(23) = 1) Then
Print #temp, f(sample, 6, 0); "  v_obscells   ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(v_obsCells(sample, celltype), 12, 2);
Next celltype
Print #temp,
End If


If (Table_Fitness_Switch(19) = 1) Then
Print #temp, f(sample, 6, 0); "  PearsonsR   ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(PearsonsR(sample, celltype), 12, t_dp(19));
Next celltype
Print #temp,
End If

If (Table_Fitness_Switch(20) = 1) Then
Print #temp, f(sample, 6, 0); "  ChiSquare   ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(ChiSquare(sample, celltype), 12, t_dp(20));
Next celltype
Print #temp,
'print #temp, "--------------------------------------------------------------------------------"
Print #temp, f(sample, 6, 0); "  Chi_inv     ";
For celltype = -2 To n_valid_Celltypes
  Print #temp, f(chiinv(n_cells(celltype)), 8, 2);
Next celltype
Print #temp,
'print #temp, "--------------------------------------------------------------------------------"
Print #temp,
End If

Print #temp,

End Sub


Sub Summarize_Cellular_Results(table_no, temp)

Dim dp(1 To 8) As Integer
'cols x rows x fitness measure x min/5%CI/ave/95%CI/max
Dim vector_out(1 To 20, 1 To 40, 1 To 8, 1 To 5) As Double
'1:1000 - Sample no.; 1:20 - col. no.; 1:40 - row. no.
Dim cell_exp_tmp(1 To 1000, 1 To 20, 1 To 40) As Double 'expected (unperturbed) count
Dim cell_obs_tmp(1 To 1000, 1 To 20, 1 To 40) As Double 'observed (perturbed) count

Dim Summary_type_label(1 To 5) As String

Dim cell_fitness_type As Integer, Cell_summary_type As Integer
Dim N_Cell_Fitness_Types As Integer
Dim N_Cell_Fitness_Switches_On As Integer, N_Cell_Summary_Switches_On As Integer
Dim celltype As Long, CT As Long
Dim sample As Integer, x As Integer, y As Integer

'Initialise printing decimal places for measure
dp(1) = 5
dp(2) = 5
dp(3) = 5
dp(4) = 5
dp(5) = 5
dp(6) = 5
dp(7) = 5
dp(8) = 5

'Initialise Summary measure type labels
Summary_type_label(1) = "Maximum"
Summary_type_label(2) = "97.5%le"
Summary_type_label(3) = "Mean   "
Summary_type_label(4) = "2.5%ile"
Summary_type_label(5) = "Minimum"

'Initialise no. of cell fitness measures
N_Cell_Fitness_Types = 8

'Find number of Cell Summary measures required (mean; min; max; 5%; 95%)
N_Cell_Summary_Switches_On = 0
For x = 1 To 5
  If (Cell_Summary_Switch(x) = 1) Then N_Cell_Summary_Switches_On = N_Cell_Summary_Switches_On + 1
Next x

'Find number of Cell Fitness measures required (exp; obs; Z; NFC; etc)
N_Cell_Fitness_Switches_On = 0
For x = 1 To N_Cell_Fitness_Types
  If (Cell_Fitness_Switch(x) = 1) Then N_Cell_Fitness_Switches_On = N_Cell_Fitness_Switches_On + 1
Next x


'For required measures, find ave/min/max/%CIs across samples
For cell_fitness_type = 1 To N_Cell_Fitness_Types

If (Cell_Fitness_Switch(cell_fitness_type) = 1) Then 'If summary output required for measure
  Select Case cell_fitness_type
  Case 1
    'Create temp. version of cell_exp without table dimension
    For sample = 1 To 1000
      For x = 1 To cols(table_no)
        For y = 1 To rows(table_no)
          cell_exp_tmp(sample, x, y) = cell_exp(table_no, sample, x, y)
        Next y
      Next x
    Next sample
    Call Calc_Cell_Summary(cell_exp_tmp, vector_out, cell_fitness_type, table_no, temp)
  Case 2
    'Create temp. version of cell_obs without table dimension
    For sample = 1 To 1000
      For x = 1 To cols(table_no)
        For y = 1 To rows(table_no)
          cell_obs_tmp(sample, x, y) = cell_obs(table_no, sample, x, y)
        Next y
      Next x
    Next sample
    Call Calc_Cell_Summary(cell_obs_tmp, vector_out, cell_fitness_type, table_no, temp)
  Case 3
    Call Calc_Cell_Summary(cell_changed, vector_out, cell_fitness_type, table_no, temp)
  Case 4
    Call Calc_Cell_Summary(cell_diff, vector_out, cell_fitness_type, table_no, temp)
  Case 5
    Call Calc_Cell_Summary(cell_Z, vector_out, cell_fitness_type, table_no, temp)
  Case 6
    Call Calc_Cell_Summary(cell_NFC, vector_out, cell_fitness_type, table_no, temp)
  Case 7
    Call Calc_Cell_Summary(cell_Zm, vector_out, cell_fitness_type, table_no, temp)
  Case 8
    Call Calc_Cell_Summary(cell_NFCm, vector_out, cell_fitness_type, table_no, temp)
  End Select
  
End If ' if measure switch is on

Next cell_fitness_type

'Write out summary results (average measure of cell fit across samples)

If (N_Cell_Summary_Switches_On > 0) And (N_Cell_Fitness_Switches_On > 0) Then 'write out cell fitness measure averages

Print #temp, "=== Table-specific, Cross-area, Cell-based report (mean); "; tablename; " ==="
Print #temp,
  
  For y = 1 To rows(table_no)
    
    For cell_fitness_type = 1 To N_Cell_Fitness_Types
      
      'Print out current fitness measure (incl. label) for all cells in row, if switch is 'on'
      If Cell_Fitness_Switch(cell_fitness_type) = 1 Then
        Print #temp, Cell_Fitness_label(cell_fitness_type); " "; Summary_type_label(3);
        For x = 1 To cols(table_no)
          Print #temp, f(vector_out(x, y, cell_fitness_type, 3), 12, dp(cell_fitness_type));
        Next x
        Print #temp, 'end current row
      End If
      
    Next cell_fitness_type
    
    Print #temp, 'put blank line between table cells

  Next y
  Print #temp, 'put extra blank line between tables
  
End If


'Write out full summary results (average and dispersion of fit across samples)
'at Tabular level

If (N_Cell_Summary_Switches_On > 0) And (N_Cell_Fitness_Switches_On > 0) Then 'write out all required cell fitness measure averages

Print #temp, "=== Table-specific, Cross-area, Cell-based report (user-requested); "; tablename; " ==="
Print #temp,
  
  For cell_fitness_type = 1 To N_Cell_Fitness_Types
    If Cell_Fitness_Switch(cell_fitness_type) = 1 Then
      For y = 1 To rows(table_no)
        For Cell_summary_type = 1 To 5
          If (Cell_Summary_Switch(Cell_summary_type) = 1) Then
            Print #temp, Cell_Fitness_label(cell_fitness_type); " "; Summary_type_label(Cell_summary_type);
            For x = 1 To cols(table_no)
              Print #temp, f(vector_out(x, y, cell_fitness_type, Cell_summary_type), 12, dp(cell_fitness_type));
            Next x
            Print #temp, 'end current row
          End If
        Next Cell_summary_type
        'Put blank line between output for each table cell if >1 measure per cell required
        If (N_Cell_Summary_Switches_On > 1) Then Print #temp,
      Next y
      'Put additional blank line between each output table
      Print #temp,
    End If
  Next cell_fitness_type


End If 'if N_Cell_Summary_Switches_On > 0

End Sub

Sub Summarize_Tabular_Results(temp)

'cell type x fitness measure x min/5%CI/ave/95%CI/max
Dim vector_out(-2 To 50, 1 To 23, 1 To 5) As Double

Dim Summary_type_label(1 To 5) As String

Dim fitness_type As Integer, Summary_type As Integer, N_Table_Fitness_Types As Integer
Dim N_Table_Fitness_Switches_On As Integer, N_Table_Summary_Switches_On As Integer
Dim celltype As Long, CT As Long, x As Integer


'Initialise Summary measure type labels
Summary_type_label(1) = "Maximum"
Summary_type_label(2) = "95%tile"
Summary_type_label(3) = "Mean   "
Summary_type_label(4) = "5%_tile"
Summary_type_label(5) = "Minimum"

'Initialise no. of table fitness measures
N_Table_Fitness_Types = 23

'Find number of Table Fitness measures required (exp; obs; Z; NFC; etc)
N_Table_Fitness_Switches_On = 0
For x = 1 To N_Table_Fitness_Types
  If (Table_Fitness_Switch(x) = 1) Then N_Table_Fitness_Switches_On = N_Table_Fitness_Switches_On + 1
Next x

'Find number of Table Summary measures required (mean; min; max; 5%; 95%)
N_Table_Summary_Switches_On = 0
For x = 1 To 5
  If (Table_Summary_Switch(x) = 1) Then _
    N_Table_Summary_Switches_On = N_Table_Summary_Switches_On + 1
Next x


'For required measures, find ave/min/max/%CIs across samples
For fitness_type = 1 To N_Table_Fitness_Types

If (Table_Fitness_Switch(fitness_type) = 1) Then 'If summary output required for measure
  Select Case fitness_type
  Case 1
    Call Calc_Table_Summary(n_changed, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 2
    Call Calc_Table_Summary(p_changed, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 3
    Call Calc_Table_Summary(max_change, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 4
    Call Calc_Table_Summary(max_p_change, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 5
    Call Calc_Table_Summary(TE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 6
    Call Calc_Table_Summary(TAE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 7
    Call Calc_Table_Summary(RAE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 8
    Call Calc_Table_Summary(SAE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 9
    Call Calc_Table_Summary(SE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 10
    Call Calc_Table_Summary(RMSE, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 11
    Call Calc_Table_Summary(SSZ, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 12
    Call Calc_Table_Summary(NFC, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 13
    Call Calc_Table_Summary(NFT, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 14
    Call Calc_Table_Summary(SSZm, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 15
    Call Calc_Table_Summary(NFCm, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 16
    Call Calc_Table_Summary(NFTm, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 17
    Call Calc_Table_Summary(D, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 18
    Call Calc_Table_Summary(CramersV, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 19
    Call Calc_Table_Summary(PearsonsR, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 20
    Call Calc_Table_Summary(ChiSquare, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 21
    Call Calc_Table_Summary(TVCC, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 22
    Call Calc_Table_Summary(v_expCells, vector_out, fitness_type, n_valid_Celltypes, temp)
  Case 23
    Call Calc_Table_Summary(v_obsCells, vector_out, fitness_type, n_valid_Celltypes, temp)
  End Select
  
End If ' if measure switch is on

Next fitness_type

'Report tabular results (if report_table_diffs flag is on)

If ((N_Table_Summary_Switches_On > 0) And _
  (N_Table_Fitness_Switches_On > 0) And (Table_Summary_Switch(3) = 1)) Then

Print #temp, "=== Table-specific, Cross-area, Table-based report (mean); "; tablename; " ==="
Print #temp,
  
  Print #temp, "                    Cell type (no. of contributing cells)"
  Print #temp, "Measure    Distrib    Marginal    Internal         All";
  For celltype = 1 To n_valid_Celltypes
    Print #temp, f(celltype_map(celltype), 12, 0);
  Next celltype
  Print #temp,

  If Table_Fitness_Switch(0) = 1 Then
    Print #temp, "frequency  Count  ";
    For celltype = -2 To n_valid_Celltypes
      Print #temp, f(n_cells(celltype), 12, 0);
    Next celltype
    Print #temp,
  End If

  For fitness_type = 1 To N_Table_Fitness_Types
    
    If Table_Fitness_Switch(fitness_type) = 1 Then
      Print #temp, Fitness_label(fitness_type); " "; Summary_type_label(3);
      For celltype = -2 To n_valid_Celltypes
        Print #temp, f(vector_out(celltype, fitness_type, 3), 12, t_dp(fitness_type));
      Next celltype
    Print #temp,
    If Count_table_in_use = 1 Then Print #temp, 'blank line between measures
    End If
    
  Next fitness_type '(of fitness)

  Print #temp, 'put blank line between tables

End If 'If N_Table_Summary_Switches_On > 0

'Write out full summary results (average and dispersion of fit across samples)
'at Tabular level [unless only mean has been requested, in which case, this has
'already been written out above...]

If (N_Table_Summary_Switches_On > 0) And (N_Table_Fitness_Switches_On > 0) Then

  If (N_Table_Summary_Switches_On > 1) Or _
    (N_Table_Summary_Switches_On = 1 And Table_Summary_Switch(3) <> 1) Then

Print #temp, "=== Table-specific, Cross-area, Table-based report (user-requested); "; tablename; " ==="
Print #temp,
  
  Print #temp, "                    Cell type (no. of contributing cells)"
  Print #temp, "Measure    Distrib    Marginal    Internal         All";
  For celltype = 1 To n_valid_Celltypes
    Print #temp, f(celltype_map(celltype), 12, 0);
  Next celltype
  Print #temp,
  
  If Table_Fitness_Switch(0) = 1 Then
    Print #temp, "frequency  Count  ";
    For celltype = -2 To n_valid_Celltypes
      Print #temp, f(n_cells(celltype), 12, 0);
    Next celltype
    Print #temp,
  End If
    
  For fitness_type = 1 To N_Table_Fitness_Types
    
    If Table_Fitness_Switch(fitness_type) = 1 Then
    
      For Summary_type = 1 To 5
        
        If (Table_Summary_Switch(Summary_type) = 1) Then
        
          Print #temp, Fitness_label(fitness_type); " "; Summary_type_label(Summary_type);
        
          For celltype = -2 To n_valid_Celltypes

            Print #temp, f(vector_out(celltype, fitness_type, Summary_type), 12, t_dp(fitness_type));

          Next celltype
          Print #temp,
      
       End If 'if summary type switched on
       
      Next Summary_type
    
  End If 'if fitness type switched on
    
  Next fitness_type 'add blank line between tables

  Print #temp,
  
End If '2+ summary types required; or summary type required is NOT mean...

End If 'if N_Table_Fitness_Switches_On > 0

End Sub
Sub Calc_Cell_Summary(vector_in, vector_out, cell_fitness_type, table_no, temp)
Dim sample As Integer, celltype As Integer
Dim CT As Integer
Dim vector_sum As Double
Dim vector(1 To 1000) As Double, vector_map(1 To 1000) As Long
Dim Comparisons As Long, Switches As Long, i As Long
Dim First As Long, Smallest As Long, Current As Long
Dim temporary As Double, temporary_map As Long
Dim Q As Double, Qvalue As Double, Qi As Long, Qf As Double
Dim x As Integer, y As Integer


  For x = 1 To cols(table_no)
  
    For y = 1 To rows(table_no)

    '1. Vectorise (and find vector total)

    vector_sum = 0#
    
    For sample = 1 To NoOfSamples

      vector(sample) = vector_in(sample, x, y)
      vector_sum = vector_sum + vector(sample)

    Next sample

    '2. Calculate vector average
    
    vector_out(x, y, cell_fitness_type, 3) = vector_sum / NoOfSamples
    
    '3. Sort vector into ascending order

    'Selection sort algorithm
    
    Comparisons = 0
    Switches = 0
    
    'Initialise vector_map
    For First = 1 To NoOfSamples
      vector_map(First) = First
    Next First
    
    
    For First = 1 To NoOfSamples - 1
      Smallest = First
      For Current = First To NoOfSamples
        If (vector(Current) < vector(Smallest)) Then Smallest = Current
      Next Current
      temporary = vector(Smallest)
      vector(Smallest) = vector(First)
      vector(First) = temporary
    Next First
  
'    Print #temp, "i", "vector_in", "vector"
'    For i = 1 To NoOfSamples
'      Print #temp, i, vector_in(i, T, x, y), vector(i)
'    Next i
    
    '4. Identify minimum
    vector_out(x, y, cell_fitness_type, 5) = vector(1)
    
    '5. Identify maximum
    vector_out(x, y, cell_fitness_type, 1) = vector(NoOfSamples)

    '6. Identify Q1 (2.5th percentile)

    Q = 1# + ((2.5 / 100#) * (NoOfSamples - 1))  'as per Excel Q = 1+(p(N-1))
    
    Qi = Int(Q) 'Integer part of Q
    Qf = Q - Int(Q) 'Fractional part of Q

    If (Qi = 0) Then
      Qvalue = vector(1)
    ElseIf (Qi = NoOfSamples) Then
      Qvalue = vector(Qi)
    Else '0<Qi<N; interpolate Qvalue
      Qvalue = vector(Qi) + ((vector(Qi + 1) - vector(Qi)) * Qf)
    End If

    vector_out(x, y, cell_fitness_type, 4) = Qvalue

    '7. Identify Q19 (97.5th percentile)

    Q = 1# + ((97.5 / 100#) * (NoOfSamples - 1))  'as per Excel Q = 1+(p(N-1))
    
    Qi = Int(Q) 'Integer part of Q
    Qf = Q - Int(Q) 'Fractional part of Q

    If (Qi = 0) Then
      Qvalue = vector(1)
    ElseIf (Qi = NoOfSamples) Then
      Qvalue = vector(Qi)
    Else '0<Qi<N; interpolate Qvalue
      Qvalue = vector(Qi) + ((vector(Qi + 1) - vector(Qi)) * Qf)
    End If

    vector_out(x, y, cell_fitness_type, 2) = Qvalue

    '8. Report results for current cell type

'    Print #temp, "CellType ", CellType
'    For i = 1 to 5
'      Print #temp, vector_out(x, y, cell_fitness_type, i)
'    Next i

    Next y 'next row

  Next x 'next column


End Sub

Sub Calc_Table_Summary(vector_in, vector_out, fitness_type, n_celltypes, temp)
Dim sample As Integer, T As Integer, celltype As Integer
Dim CT As Integer
Dim vector_sum As Double
Dim vector(1 To 1000) As Double, vector_map(1 To 1000) As Long
Dim Comparisons As Long, Switches As Long, i As Long
Dim First As Long, Smallest As Long, Current As Long
Dim temporary As Double, temporary_map As Long
Dim Q As Double, Qvalue As Double, Qi As Long, Qf As Double


  'Loop for each unique 'cell types' in current table
  For celltype = -2 To n_celltypes
    
    '1. Vectorise (and find vector total)

    vector_sum = 0#
    
    For sample = 1 To NoOfSamples

      vector(sample) = vector_in(sample, celltype)
      vector_sum = vector_sum + vector(sample)

    Next sample

    '2. Calculate vector average
    
    vector_out(celltype, fitness_type, 3) = vector_sum / NoOfSamples
    
    '3. Sort vector into ascending order

    'Selection sort algorithm
    
    Comparisons = 0
    Switches = 0
    
    'Initialise vector_map
    For First = 1 To NoOfSamples
      vector_map(First) = First
    Next First
    
    
    For First = 1 To NoOfSamples - 1
      Smallest = First
      For Current = First To NoOfSamples
        If (vector(Current) < vector(Smallest)) Then Smallest = Current
      Next Current
      temporary = vector(Smallest)
      vector(Smallest) = vector(First)
      vector(First) = temporary
    Next First
  
'    Print #temp, "i", "vector_in", "vector"
'    For i = 1 To NoOfSamples
'      Print #temp, i, vector_in(i, CellType), vector(i)
'    Next i
    
    '4. Identify minimum
    vector_out(celltype, fitness_type, 5) = vector(1)
    
    '5. Identify maximum
    vector_out(celltype, fitness_type, 1) = vector(NoOfSamples)

    '6. Find value of 2.5th percentile

    Q = 1# + ((2.5 / 100#) * (NoOfSamples - 1))   'as per Excel Q = 1+(p(N-1))
    
    Qi = Int(Q) 'Integer part of Q
    Qf = Q - Int(Q) 'Fractional part of Q

    If (Qi = 0) Then 'Sample size = 1; set Q value to only observed sample value
      Qvalue = vector(1)
    ElseIf (Qi = NoOfSamples) Then 'in cases with few observations/categories,
                                  '95/100*(N+1) can be > N, in which case
      Qvalue = vector(Qi)         'set Qvalue to largest observed sample value
    Else '0<Qi<N; interpolate Qvalue
      Qvalue = vector(Qi) + ((vector(Qi + 1) - vector(Qi)) * Qf)
    End If

    vector_out(celltype, fitness_type, 4) = Qvalue

    '7. Find value of 97.5th percentile

    Q = 1# + ((97.5 / 100#) * (NoOfSamples - 1))   'as per Excel Q = 1+(p(N-1))
    
    Qi = Int(Q) 'Integer part of Q
    Qf = Q - Int(Q) 'Fractional part of Q

    If (Qi = 0) Then 'Sample size = 1; set Q value to only observed sample value
      Qvalue = vector(1)
    ElseIf (Qi = NoOfSamples) Then 'in cases with few observations/categories,
                                  '95/100*(N+1) can be > N, in which case
      Qvalue = vector(Qi)         'set Qvalue to largest observed sample value
    Else '0<Qi<N; interpolate Qvalue
      Qvalue = vector(Qi) + ((vector(Qi + 1) - vector(Qi)) * Qf)
    End If

    vector_out(celltype, fitness_type, 2) = Qvalue

    '8. Report results for current cell type

'    Print #temp, "CellType ", CellType
'    For i = 1 to 5
'      Print #temp, vector_out(CellType, fitness_type, i)
'    Next i

  Next celltype 'next unique cell type in table

End Sub
Sub Report_Cross_Table_Sample_Summary(temp)
Dim sample As Integer, x As Integer
Dim fitness_type As Integer
Dim celltype As Integer, CellType_Name(-2 To 200) As String

'Initialise CellType_Name labels
CellType_Name(-2) = "Marginal "
CellType_Name(-1) = "Internal "
CellType_Name(0) = "All "
For celltype = 1 To MaxContribCells_tot
  CellType_Name(celltype) = celltype
Next celltype


Print #temp, "=== Cross-table, Area-specific, Table-based report ==="
Print #temp,

'Report frequency of each cell type, if required
If Table_Fitness_Switch(0) = 1 Then
  Print #temp, "                    Cell type (no. of contributing cells)"
  Print #temp, "   Measure  Sample";
  For celltype = -2 To MaxContribCells_tot
  If (n_cells_tot(celltype) > 0) Then Print #temp, f(CellType_Name(celltype), 12, 0);
  Next celltype
  Print #temp,
  Print #temp, "   frequency  All ";
  For celltype = -2 To MaxContribCells_tot
    If (n_cells_tot(celltype) > 0) Then _
      Print #temp, f(n_cells_tot(celltype), 12, 0);
  Next celltype
  Print #temp,
  Print #temp,
End If

'Report other user-requested cross-table measures
For fitness_type = 1 To 23

If Table_Fitness_Switch(fitness_type) = 1 Then

Print #temp, "                    Cell type (no. of contributing cells)"
Print #temp, "   Measure  Sample";
For celltype = -2 To MaxContribCells_tot
  If (n_cells_tot(celltype) > 0) Then Print #temp, f(CellType_Name(celltype), 12, 0);
Next celltype
Print #temp,

Select Case fitness_type
  Case 1 'n_changed_tot
   Call print_sample_specific_table_summary(n_changed_tot, fitness_type, temp)
  Case 2 'p_changed_tot
    Call print_sample_specific_table_summary(p_changed_tot, fitness_type, temp)
  Case 3 'max_change_tot
   Call print_sample_specific_table_summary(max_change_tot, fitness_type, temp)
  Case 4 'max_p_change_tot
    Call print_sample_specific_table_summary(max_p_change_tot, fitness_type, temp)
  Case 5 'TotalError
    Call print_sample_specific_table_summary(TE_tot, fitness_type, temp)
  Case 6 'TAE
    Call print_sample_specific_table_summary(TAE_tot, fitness_type, temp)
  Case 7 'RAE
    Call print_sample_specific_table_summary(RAE_tot, fitness_type, temp)
  Case 8 'SAE
    Call print_sample_specific_table_summary(SAE_tot, fitness_type, temp)
  Case 9 'Sq_Error
    Call print_sample_specific_table_summary(SE_tot, fitness_type, temp)
  Case 10 'RMSE
    Call print_sample_specific_table_summary(RMSE_tot, fitness_type, temp)
  Case 11 'SSZ
    Call print_sample_specific_table_summary(SSZ_tot, fitness_type, temp)
  Case 12 'NFC
    Call print_sample_specific_table_summary(NFC_tot, fitness_type, temp)
  Case 13 'NFT
    Call print_sample_specific_table_summary(NFT_tot, fitness_type, temp)
  Case 14 'SSZm
    Call print_sample_specific_table_summary(SSZm_tot, fitness_type, temp)
  Case 15 'NFCm
    Call print_sample_specific_table_summary(NFCm_tot, fitness_type, temp)
  Case 16 'NFTm
    Call print_sample_specific_table_summary(NFTm_tot, fitness_type, temp)
  Case 17 'GibsonsD
    Call print_sample_specific_table_summary(D_tot, fitness_type, temp)
  Case 18 'CramersV
    Call print_sample_specific_table_summary(CramersV_tot, fitness_type, temp)
  Case 19 'PearsonsR
    Call print_sample_specific_table_summary(PearsonsR_tot, fitness_type, temp)
  Case 20 'ChiSquare
    If sample = 1 Then
      'print chisquare critical values
      'print #temp, "--------------------------------------------------------------------------------"
      Print #temp, f(sample, 6, 0); "  Chi_inv     ";
      For celltype = -2 To n_valid_Celltypes
        Print #temp, f(chiinv(n_cells_tot(celltype)), 8, 2);
      Next celltype
      Print #temp,
      'print #temp, "--------------------------------------------------------------------------------"
      Print #temp,
    End If 'if first sample
    Call print_sample_specific_table_summary(ChiSquare_tot, fitness_type, temp)
  Case 21 'TVCC
    Call print_sample_specific_table_summary(TVCC_tot, fitness_type, temp)
  Case 22 'v_expCells
    Call print_sample_specific_table_summary(v_expCells_tot, fitness_type, temp)
  Case 23 'v_obsCells
    Call print_sample_specific_table_summary(v_obsCells_tot, fitness_type, temp)
End Select

  
End If 'if fitness_switch=1 [on]

Next fitness_type
Print #temp,

End Sub
Sub print_sample_specific_table_summary(vector_in, fitness_type, temp)
Dim sample As Integer, x As Integer

For sample = 1 To NoOfSamples
  Print #temp, f(Fitness_label(fitness_type), 12, 0);
  Print #temp, f(sample, 6, 0);
  For x = -2 To MaxContribCells_tot
    If n_cells_tot(x) > 0 Then 'if at least one cell of this type...
      Print #temp, f(vector_in(sample, x), 12, t_dp(fitness_type));
    End If
  Next x
  Print #temp,
Next sample
Print #temp,

End Sub

Sub Report_Cross_Table_Overall_Summary(temp)

'cell type across all tables (mapped) x fitness measure x min/5%CI/ave/95%CI/max
Dim vector_out(-2 To 200, 1 To 23, 1 To 5) As Double

Dim Summary_type_label(1 To 5) As String

Dim fitness_type As Integer, Summary_type As Integer, N_Table_Fitness_Types As Integer
Dim N_Table_Fitness_Switches_On As Integer, N_Table_Summary_Switches_On As Integer
Dim celltype As Long, CT As Long, x As Integer
Dim CellType_Name(-2 To 200) As String '-2:200 = cell type across all tables (mapped)

'Initialise CellType_Name labels
CellType_Name(-2) = "Marginal "
CellType_Name(-1) = "Internal "
CellType_Name(0) = "All "
For celltype = 1 To MaxContribCells_tot
  CellType_Name(celltype) = celltype
Next celltype

'Initialise Summary measure type labels
Summary_type_label(1) = "Maximum"
Summary_type_label(2) = "95%tile"
Summary_type_label(3) = "Mean   "
Summary_type_label(4) = "5%_tile"
Summary_type_label(5) = "Minimum"

'Initialise no. of table fitness measures
N_Table_Fitness_Types = 23

'Find number of Table Fitness measures required (exp; obs; Z; NFC; etc)
N_Table_Fitness_Switches_On = 0
For x = 1 To N_Table_Fitness_Types
  If (Table_Fitness_Switch(x) = 1) Then N_Table_Fitness_Switches_On = N_Table_Fitness_Switches_On + 1
Next x

'Find number of Table Summary measures required (mean; min; max; 5%; 95%)
N_Table_Summary_Switches_On = 0
For x = 1 To 5
  If (Table_Summary_Switch(x) = 1) Then N_Table_Summary_Switches_On = N_Table_Summary_Switches_On + 1
Next x

'For required measures, find ave/min/max/%CIs across samples
For fitness_type = 1 To N_Table_Fitness_Types

If (Table_Fitness_Switch(fitness_type) = 1) Then 'If summary output required for measure
  Select Case fitness_type
  Case 1
    Call Calc_Table_Summary(n_changed_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 2
    Call Calc_Table_Summary(p_changed_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 3
    Call Calc_Table_Summary(max_change_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 4
    Call Calc_Table_Summary(max_p_change_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 5
    Call Calc_Table_Summary(TE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 6
    Call Calc_Table_Summary(TAE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 7
    Call Calc_Table_Summary(RAE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 8
    Call Calc_Table_Summary(SAE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 9
    Call Calc_Table_Summary(SE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 10
    Call Calc_Table_Summary(RMSE_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 11
    Call Calc_Table_Summary(SSZ_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 12
    Call Calc_Table_Summary(NFC_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 13
    Call Calc_Table_Summary(NFT_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 14
    Call Calc_Table_Summary(SSZm_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 15
    Call Calc_Table_Summary(NFCm_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 16
    Call Calc_Table_Summary(NFTm_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 17
    Call Calc_Table_Summary(D_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 18
    Call Calc_Table_Summary(CramersV_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 19
    Call Calc_Table_Summary(PearsonsR_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 20
    Call Calc_Table_Summary(ChiSquare_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 21
    Call Calc_Table_Summary(TVCC_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 22
    Call Calc_Table_Summary(v_expCells_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  Case 23
    Call Calc_Table_Summary(v_obsCells_tot, vector_out, fitness_type, MaxContribCells_tot, temp)
  End Select
End If ' if measure switch is on

Next fitness_type

'Write out summary results (average measure of fit across samples)

If (N_Table_Summary_Switches_On > 0) And _
  ((N_Table_Fitness_Switches_On > 0) And (Table_Summary_Switch(3) = 1)) Then

  Print #temp, "=== Cross-table, Cross-area, Table-based report (mean) ==="
  Print #temp,

  Print #temp, "                    Cell type (no. of contributing cells)"
  Print #temp, "Measure    Distrib";
  For celltype = -2 To MaxContribCells_tot
    If (n_cells_tot(celltype) > 0) Then Print #temp, f(CellType_Name(celltype), 12, 0);
  Next celltype
  Print #temp,

  If Table_Fitness_Switch(0) = 1 Then
    Print #temp, "frequency  Count  ";
    For celltype = -2 To MaxContribCells_tot
      If (n_cells_tot(celltype) > 0) Then _
        Print #temp, f(n_cells_tot(celltype), 12, 0);
    Next celltype
    Print #temp,
  End If

  For fitness_type = 1 To N_Table_Fitness_Types
    
    If Table_Fitness_Switch(fitness_type) = 1 Then
      Print #temp, Fitness_label(fitness_type); " "; Summary_type_label(3);
      For celltype = -2 To MaxContribCells_tot
        If (n_cells_tot(celltype) > 0) Then _
          Print #temp, f(vector_out(celltype, fitness_type, 3), 12, t_dp(fitness_type));
      Next celltype
    Print #temp,
    End If
    
  Next fitness_type '(of fitness)

  Print #temp, 'put blank line between tables


End If 'If N_Table_Summary_Switches_On > 0

'Write out full summary results (average and dispersion of fit across samples)
'at Tabular level

If (N_Table_Summary_Switches_On > 0) And (N_Table_Fitness_Switches_On > 0) Then

  If (N_Table_Summary_Switches_On > 1) Or _
    (N_Table_Summary_Switches_On = 1 And Table_Summary_Switch(3) <> 1) Then

  Print #temp, "=== Cross-table, Cross-area, Table-based report (user requested) ==="
  Print #temp,
  
  Print #temp, "                    Cell type (no. of contributing cells)"
  Print #temp, "Measure    Distrib";
  For celltype = -2 To MaxContribCells_tot
    If (n_cells_tot(celltype) > 0) Then Print #temp, f(CellType_Name(celltype), 12, 0);
  Next celltype
  Print #temp,
  
  If Table_Fitness_Switch(0) = 1 Then
    Print #temp, "frequency  Count  ";
    For celltype = -2 To MaxContribCells_tot
      If (n_cells_tot(celltype) > 0) Then _
        Print #temp, f(n_cells_tot(celltype), 12, 0);
    Next celltype
    Print #temp,
  End If
  
  For fitness_type = 1 To N_Table_Fitness_Types
    
    If Table_Fitness_Switch(fitness_type) = 1 Then
    
      For Summary_type = 1 To 5
        
        If (Table_Summary_Switch(Summary_type) = 1) Then
        
          Print #temp, Fitness_label(fitness_type); " "; Summary_type_label(Summary_type);
        
          For celltype = -2 To MaxContribCells_tot

            If n_cells_tot(celltype) > 0 Then _
              Print #temp, f(vector_out(celltype, fitness_type, Summary_type), 12, t_dp(fitness_type));

          Next celltype
          Print #temp,
          
       End If 'if summary type switched on
       
      Next Summary_type
    
    End If 'if fitness type switched on
  
  Next fitness_type

  Print #temp,
  
  End If '2+ summary types required; or summary type requires is NOT mean...
  
End If 'if N_Table_Fitness_Switches_On > 0


End Sub


Sub check_rankings(table_no, percentage_name, no_of_classes, first_quantile, temp)

'arrays indexed by sample no.
Dim obs_pctg_vector(1 To 1000) As Double, exp_pctg_vector(1 To 1000) As Double
Dim obs_rank(1 To 1000) As Double, exp_rank(1 To 1000) As Double
Dim obs_class(1 To 1000) As Double, exp_class(1 To 1000) As Double
'arrays indexed by class (percentile) no.
Dim lower_bound(1 To 100) As Integer, upper_bound(1 To 100) As Integer
Dim CorrectClass(1 To 20, 1 To 40, 0 To 100) As Integer
Dim NearClass(1 To 20, 1 To 40, 0 To 100) As Integer
Dim no_in_class(1 To 20, 1 To 40, 0 To 100) As Integer
Dim Percentile(1 To 100) As Double
'arrays indexed by no. of rows and cols in current table
Dim CorrectRank(1 To 20, 1 To 40) As Integer
Dim NearRank(1 To 20, 1 To 40) As Integer
'scratch vars
Dim Class As Integer
Dim Q As Double, x As Integer, y As Integer, s As Integer
Dim report_ranking As Integer



'Identify class cut-points (Qx - Xth percentile)
'(i) upper-bound
For Class = 1 To no_of_classes
  Percentile(Class) = (100 / no_of_classes) * Class
  Q = 1# + ((Percentile(Class) / 100#) * (NoOfSamples - 1))  'as per Excel Q = 1+(p(N-1))
  upper_bound(Class) = Int(Q) 'Integer part of Q
Next Class
'(ii) lower-bound
lower_bound(1) = 1
For Class = 2 To no_of_classes
  lower_bound(Class) = upper_bound(Class - 1) + 1
Next Class


'Go through each percentage in turn
For y = 1 To rows(table_no)
  For x = 1 To cols(table_no)
  
    'Create temporary vector of current percentage
    For s = 1 To NoOfSamples
      obs_pctg_vector(s) = cell_obs(table_no, s, x, y)
      exp_pctg_vector(s) = cell_exp(table_no, s, x, y)
    Next s
    
    report_ranking = 0
    
    'Device to report ranking of selected variables
    '[needed for check ranking.xls ]
    'If (x = 17 And y = 1) Then
    '  report_ranking = 1
    'End If
    
    'Sort vectors into rank order (1=low; NoOfSamples=High)
    Call sort(exp_pctg_vector, exp_rank, report_ranking, temp)
    Call sort(obs_pctg_vector, obs_rank, report_ranking, temp)
        
    'Identify percentile class sample falls within
    For s = 1 To NoOfSamples
      For Class = 1 To no_of_classes
        If obs_rank(s) >= lower_bound(Class) Then obs_class(s) = Class
        If exp_rank(s) >= lower_bound(Class) Then exp_class(s) = Class
      Next Class
    Next s

    'Check accuracy of ranking after SDC
    CorrectRank(x, y) = 0
    NearRank(x, y) = 0
    For Class = 0 To no_of_classes
      CorrectClass(x, y, Class) = 0
      NearClass(x, y, Class) = 0
    Next Class
    For s = 1 To NoOfSamples
      'Check to see if rank exactly right
      If obs_rank(s) = exp_rank(s) Then CorrectRank(x, y) = CorrectRank(x, y) + 1
      'Count no. in expected (unperturbed) class
      no_in_class(x, y, 0) = no_in_class(x, y, 0) + 1
      no_in_class(x, y, exp_class(s)) = no_in_class(x, y, exp_class(s)) + 1
      'check to see if perturbed sample ranked in same class
      If obs_class(s) = exp_class(s) Then
        CorrectClass(x, y, exp_class(s)) = CorrectClass(x, y, exp_class(s)) + 1
        CorrectClass(x, y, 0) = CorrectClass(x, y, 0) + 1
      End If
      'check to see if perturbed sample ranked in same or neighbouring class
      If Abs(obs_class(s) - exp_class(s)) <= 1 Then
        NearClass(x, y, exp_class(s)) = NearClass(x, y, exp_class(s)) + 1
        NearClass(x, y, 0) = NearClass(x, y, 0) + 1
      End If
    Next s

  Next x 'next col (%)

Next y 'next row (%)

'Report Correct Rank (if requested)
If (report_ranking_flag(1) = 1 And first_quantile = 1) Then
  Print #temp, "=== Correct Rank; "; tablename; " ==="
  Print #temp,
  Call report_rankings(percentage_name, CorrectRank, CorrectClass, _
    no_in_class, no_of_classes, 1, table_no, temp)
End If

'Report bounds [if Correct and/or Correct/Neighbouring Class reports required]
If (report_ranking_flag(2) = 1 Or report_ranking_flag(3) = 1) Then
  Print #temp, "=== Quantile boundaries ("; no_of_classes; " classes); "; tablename; " ==="
  Print #temp,
  For Class = 1 To no_of_classes
  Print #temp, "Percentile: "; Percentile(Class); " class: "; Class; _
      " Lower-bound: "; lower_bound(Class); " Upper-bound: "; upper_bound(Class)
  Next Class
  Print #temp,
End If 'if individual or overall class reports required

'Report Correct Classes (if required)
If report_ranking_flag(2) = 1 Then
  Print #temp, "=== Correct Class ("; no_of_classes; " quantiles); "; tablename; " ==="
  Print #temp,
  Call report_rankings(percentage_name, CorrectRank, CorrectClass, _
    no_in_class, no_of_classes, 2, table_no, temp)
End If

'Report Correct/Neigbouring classes (if required)
If report_ranking_flag(3) = 1 Then
  Print #temp, "=== Correct/Neighbouring class ("; no_of_classes; " quantiles); "; tablename; " ==="
  Print #temp,
  Call report_rankings(percentage_name, CorrectRank, NearClass, _
    no_in_class, no_of_classes, 3, table_no, temp)
End If

End Sub


Sub report_rankings(percentage_name, CorrectRank, C_or_N_Class, no_in_class, _
  no_of_classes, report_type, table_no, temp)

Dim x As Integer, y As Integer, Class As Integer, report_name As String

If report_type = 1 Then

For y = 1 To rows(table_no)
  'Report accuracy of rankings for all perecentages in current row
  Print #temp, " ";
  For x = 1 To cols(table_no)
    Print #temp, percentage_name(x, y); " "; percentage_name(x, y); " "; percentage_name(x, y); " ";
  Next x
  Print #temp,
  Print #temp, " ";
  For x = 1 To cols(table_no)
    Print #temp, "CorrectRank "; "Samples "; "%_correct ";
  Next x
  Print #temp,
  For x = 1 To cols(table_no)
    Print #temp, CorrectRank(x, y), NoOfSamples, f(CorrectRank(x, y) / NoOfSamples * 100, 8, 2);
  Next x
  Print #temp,
  Print #temp,
Next y

End If

If (report_type = 2) Then ' if actual class correct
  report_name = "Correct_Class "
ElseIf (report_type = 3) Then ' if actual or neighbouring class correct
  report_name = "Near_Class "
End If

If (report_type > 1) Then

For y = 1 To rows(table_no)

  Print #temp, " Percentage ";
  For x = 1 To cols(table_no)
    Print #temp, percentage_name(x, y); " "; percentage_name(x, y); " "; percentage_name(x, y); " ";
  Next x
  Print #temp,
  
  Print #temp, " Class ";
  For x = 1 To cols(table_no)
    Print #temp, report_name; "no._in_class "; "%_Correct    ";
  Next x
  Print #temp,

  For Class = 1 To no_of_classes
    Print #temp, Class;
    For x = 1 To cols(table_no)
      If no_in_class(x, y, Class) > 0 Then
        Print #temp, C_or_N_Class(x, y, Class); no_in_class(x, y, Class); _
          f(C_or_N_Class(x, y, Class) / no_in_class(x, y, Class) * 100, 8, 2);
      ElseIf C_or_N_Class(x, y, Class) = 0 Then
        Print #temp, C_or_N_Class(x, y, Class); no_in_class(x, y, Class); "  0";
      Else
        Print #temp, C_or_N_Class(x, y, Class); no_in_class(x, y, Class); " -9";
      End If
    Next x
    Print #temp,
  Next Class
  Print #temp,
  Print #temp,

  Print #temp, " ";
  Print #temp, "All   ";
  For x = 1 To cols(table_no)
    Print #temp, report_name; "no._in_sample"; " %_Correct ";
  Next x
  Print #temp,
  Print #temp, " classes ";
  For x = 1 To cols(table_no)
    Print #temp, C_or_N_Class(x, y, 0); no_in_class(x, y, 0); _
      f(C_or_N_Class(x, y, 0) / NoOfSamples * 100, 8, 2);
  Next x
  Print #temp,
  
  Print #temp,

Next y 'next table row

End If 'if report_type > 1

End Sub

Sub sort(vector_in, vector_rank, report_ranking, temp)

Dim First As Long, Smallest As Long, Current As Long, lowest_rank As Integer
Dim temporary As Double, i As Integer
Dim vector_copy(1 To 1000) As Double 'Should be delcared same as vector_in
Dim vector_map(1 To 1000) As Integer

'1. Initialise vector_map and vector_copy
For i = 1 To NoOfSamples
  vector_map(i) = i
  vector_copy(i) = vector_in(i)
Next i

'2. Sort vector into ascending order
'   [Uses 'selection sort' algorithm]
    
For First = 1 To NoOfSamples - 1
  Smallest = First
  For Current = First To NoOfSamples
    If (vector_copy(Current) < vector_copy(Smallest)) Then Smallest = Current
  Next Current
  'swap records in copy of vector_in
  temporary = vector_copy(Smallest)
  vector_copy(Smallest) = vector_copy(First)
  vector_copy(First) = temporary
  'Swap records in vector_map
  temporary = vector_map(Smallest)
  vector_map(Smallest) = vector_map(First)
  vector_map(First) = temporary
Next First

'3. Assign initial rank to each sample
For i = 1 To NoOfSamples
  vector_rank(vector_map(i)) = i
Next i

If report_ranking = 1 Then
  Print #temp, " i", "vector_in", "vector_map", "vector_copy", _
    "vector_in(vector_map(i))", "vector_rank"
  For i = 1 To NoOfSamples
    Print #temp, i, vector_in(i), vector_map(i), vector_copy(i), _
      vector_in(vector_map(i)), vector_rank(i)
  Next i
  Print #temp,
End If
    
'4. At this stage, observations with same value have unique ranks, based on
'   orginal sample order; the following reassigns all 2nd and subsequent duplicates
'   the same (lowest) rank as the first ranked instance.

lowest_rank = 1
For i = 2 To NoOfSamples
  If (vector_in(vector_map(i)) <> vector_in(vector_map(i - 1))) Then
    lowest_rank = vector_rank(vector_map(i))
  End If
  vector_rank(vector_map(i)) = lowest_rank
Next i
    
If report_ranking = 1 Then
  Print #temp, " i", "vector_in", "vector_map", "vector_copy", _
    "vector_in(vector_map(i))", "vector_rank"
  For i = 1 To NoOfSamples
    Print #temp, i, vector_in(i), vector_map(i), vector_copy(i), _
      vector_in(vector_map(i)), vector_rank(i)
  Next i
  Print #temp,
End If
    
    
End Sub

